# -*- coding: utf-8 -*-
"""
Created on Thu Sep 24 09:24:31 2020

@author: Amirh
"""

from ..utils._base_feature_selector import BaseFeatureSelector
from ..functions import ff
import numpy as np
import random
from sklearn.base import clone
import time
from scipy.stats import iqr


class _AntColony():    
    def __init__(self, estimator,cv,scorer, nf, fitness, weight,
                 x,y,x_test= None, y_test = None,
                 MaxIt = 1000, n_ant = 10, Q=1, tau0=1, alpha=1, beta=1,
                 rho=0.5, random_state=None,verbose = True):
        
        self.nf = nf
        self.estimator = estimator
        self.cv = cv
        self.scorer = scorer
        self.x = x
        self.y = y
        self.x_test = x_test
        self.y_test = y_test
        self.fitness = fitness
        self.weight = weight    
        self.MaxIt = MaxIt
        self.n_ant = n_ant
        
        # ACO params
        self.Q = Q #pheromone strength
        self.tau0 = tau0 #initial phromone
        self.alpha = alpha #Phromone Exponential Weight
        self.beta = beta #Heuristic Exponential Weight
        self.rho = rho #Pheromone Evaporation rate
        self.n_var = self.x.shape[1]
        self.verbose = verbose
        self.random_state = random_state
        
        random.seed(self.random_state)        
        np.random.seed(self.random_state)
        
    def CreateRandomSolution(self):
        #Function producing random permutation
        return np.random.permutation(self.n_var)
        
    def RouletteWheelSelection(self,p):
        r = np.random.uniform(0, 1)
        c = np.cumsum(p)
        i = np.where(r <= c)[0][0]
        return i    
    
    def FeatureRanker(self,all_agents):
        agents = [agent.tour for agent in all_agents]
        fits = [agent.fit for agent in all_agents]
        percs = [agent.perc for agent in all_agents]
        
        if self.weight == 1:
            fits_first = np.percentile(fits,75)
            fits_second = np.percentile(fits,100)        
        elif self.weight == -1:
            fits_first = np.percentile(fits,0)
            fits_second = np.percentile(fits,25)        
            
        indices = np.where(np.logical_and((fits >= fits_first), (fits <= fits_second)))[0]
        
        top_agents = np.array(agents)[indices]
        top_perc = np.array(percs)[indices]
        return top_agents,top_perc
    
    def generate(self):
            
        eta = np.ones((self.n_var,self.n_var)) #heuristic information matrix
        tau = self.tau0*np.ones((self.n_var,self.n_var)) # phromone matrix
        
        self.best_fits = np.zeros(self.MaxIt)  # array holding best fits
        
        #creating the colony
        Ant = type("Ant", (object,), {})        
        colony = [Ant() for i in range(self.n_ant)]       
                
        # setting up a variable to store all agents 
        all_agents = []        
        
        # best ant 
        self.BestSol = Ant()
        
        if self.weight == 1:
            self.BestSol.fit = -np.inf
        elif self.weight == -1:
            self.BestSol.fit = np.inf
        
        for it in range(self.MaxIt):
            for k in range(self.n_ant):
                
                # you want to pay more attention to this part
                # initializing the first place to start at a random position
                colony[k].tour = [random.randint(0,self.n_var-1)]
                
                # building up on that intitial guess
                for l in np.arange(1,self.nf):
                    
                    i = colony[k].tour[-1]

                    P = np.power(np.power(np.power(tau[i,:],self.alpha),eta[i,:]),self.beta)
                    
                    
                    P[colony[k].tour] = 0
                    
                    P = P/sum(P)
                    
                    j = self.RouletteWheelSelection(P)
        
                    colony[k].tour = np.hstack([colony[k].tour,j])
                
                
                if (self.x_test is not None) and (self.y_test is not None):
                    colony[k].fit, colony[k].perc = self.fitness.calculate_fitness(self.estimator,
                          self.cv,
                          self.scorer,
                          self.x[:,colony[k].tour],
                          self.y,
                          x_test=self.x_test[:,colony[k].tour],
                          y_test=self.y_test)
                    
                else:
                    colony[k].fit, colony[k].perc = self.fitness.calculate_fitness(self.estimator,
                          self.cv,
                          self.scorer,
                          self.x[:,colony[k].tour],
                          self.y)
                
                # update the best ant
                if self.weight == 1 and colony[k].fit > self.BestSol.fit:
                    self.BestSol.tour = colony[k].tour
                    self.BestSol.fit = colony[k].fit
                    self.BestSol.perc = colony[k].perc
                    
                elif self.weight == -1 and colony[k].fit < self.BestSol.fit:
                    self.BestSol.tour = colony[k].tour
                    self.BestSol.fit = colony[k].fit
                    self.BestSol.perc = colony[k].perc
                    
                
            #update phromones
            for k in range(self.n_ant):
                
                tour = colony[k].tour
                tour = np.hstack([tour,tour[0]])
                
                for l in range(self.nf):
                    i = tour[l]
                    j = tour[l+1]
                    
                    tau[i,j] = tau[i,j] + self.Q/colony[k].fit
            
            # evaporate
            tau = (1-self.rho)*tau
            
            # best cost
            self.best_fits[it] = self.BestSol.fit
                        
            # updating all agents
            all_agents.extend(colony)
            
            if self.verbose:
                print("Iter: {} Features: {}  Fitness_score: {:.4} ".format(it+1, self.BestSol.tour[:self.nf],self.BestSol.fit))                        
        
        # getting top_agents
        tops = self.FeatureRanker(all_agents) 
        
        return np.array([self.BestSol.tour, self.BestSol.fit]), self.best_fits, tops
    

    
class ACO(BaseFeatureSelector):
    def __init__(self, model, n_f, weight, scoring, n_iter = 100, n_ant = 10, cv=None,
                 Q=1, tau0=1, alpha=1, beta=2, rho=0.5, 
                 verbose= True, random_state=None,**kwargs):
        """
        Ant Colony Optimization or ACO, is another nature inspired 
        optimization algorithm, which was introduced by M. Dorigo in 
        1992. In ACO, ants or agents of the model move around the 
        solution space randomly and remember the paths taken before 
        until they reach convergence. Flexibility of this algorithm
        enables us to employ it for feature selection.        

        Parameters
        ----------
        model : class
            Instantiated Sklearn regression or classification estimator.
        n_f : int
            Number of features needed to be extracted.
        weight : int
            Maximization or minimization objective, for maximization use +1
            and for mimization use -1.
        scoring : callable
            Callable sklearn score or customized score function that takes in y_pred and y_true
         as arguments
        n_iter : int, optional
            Maximum number of iterations. The default is 100.
        n_ant : int, optional
            Number of ants or agents in the optimzation model. The default is 10.
        cv : class, optional
            Instantiated sklearn cross-validation class. The default is None.
        Q : float, optional
            The pheromone intensity 𝑄, represents the total pheromonea vailable.
            This parameter affects the convergence speed of the ACO. A large 𝑄
            results in a highly concentrated pheromone, thus possibility of 
            converging to local optima. A small 𝑄 will results in possibility
            of attaining global optima, but slowers optimization speed. 
            The default is 1.
        tau0 : float, optional
            Initial pheromone intensity. This could be equal to Q. The default 
            is 1.
        alpha : float, optional
            The information elicitation factor 𝛼, addresses accumulation 
            of pheromone regarding path selection. 𝛼 should be [0,1] with an 
            emphasis on lower end of the range ([0,0.5]). If 𝛼 is large, the
             ants tend to choose the same path, chosen by preceding 
            ants, resulting in stronger cooperation among the ants and higher 
            convergence speed and higher possibility of falling into local 
            optimum. If 𝛼 is small,the convergence speed of the ACA is slowed 
            down, and ants can have a more global search to find optimum. 
            The default is 1.
        beta : float, optional
            The meta-heuristic factor 𝛽, reflects the importance of the 
            heuristic information with regard to the ants’ path selection. 𝛽
            is always larger than 1 and smaller than 5. If 𝛽 is large, there
            is a higher probability for the algorithm to behave like a 
            greedy algorithm. If 𝛽 is small, the heuristic information has no
             effect on path selection, which leads ACA to fall into local 
             optimum. The default is 2.
        rho : float, optional
            Pheromone Evaporation Coefficient 𝜌. Rrepresents the degree of 
            pheromone evaporation, reflecting the amount of mutual(social) 
            influence among ants. 𝜌 should be [0,1] with an emphasis on lower
            end of the range ([0,0.5]). If 𝜌 is small, the global search ability 
            will be reduced. If 𝜌 is too large, it will improve the global 
            search and higher chance of finding the global optimum. 
            The default is 0.5.
        verbose : bool, optional
            Wether to print out progress messages. The default is True.
        random_state : int, optional
            Determines the random state for random number generation, 
            cross-validation, and classifier/regression model setting. 
            The default is None.

        Returns
        -------
        Instantiated Optimziation model class.

        References:
            Li, P., & Zhu, H. (2016). Parameter selection for ant colony 
            algorithm based on bacterial foraging algorithm. Mathematical 
            Problems in Engineering, 2016.

			Siemiński, A. (2013). Ant colony optimization parameter evaluation.
			In Multimedia and Internet Systems: Theory and Practice 
			(pp. 143-153). Springer, Berlin, Heidelberg.            

        """
        
        super().__init__(scoring=scoring, n_f = n_f,**kwargs)                
        # =============================================================================
        # checking the input arguments
        # =============================================================================
        if n_f != int(n_f):
            raise ValueError("n_f should be an integer")
        if weight != int(weight):
            raise ValueError("wieght should be a tuple of either -1 or +1")         
        try:
            self.fitness = ff
        except:
            raise ImportError("Cannot find/import ff.py defined")        

        if n_ant != int(n_ant):
            raise ValueError("n_ant should be an integer")
        if n_iter != int(n_iter):
            raise ValueError("n_iter should be an integer")

        self.model = model
        self.n_f = n_f
        self.cv = cv      
        self.weight = weight     
        self.scoring = scoring             
        self.n_ant = n_ant
        self.n_iter = n_iter  
        self.Q = Q
        self.tau0 = tau0 #initial phromone
        self.alpha = alpha #Phromone Exponential Weight
        self.beta = beta #Heuristic Exponential Weight
        self.rho = rho #Evaporation Rate   
        
        self.verbose = verbose
        self.random_state = random_state
        
        try:
            self.cv.random_state = self.random_state
            self.cv.shuffle = True
        except:
            pass
        
        try:
            self.model.random_state = self.random_state
        except:
            pass



    def fit(self,x,y,decor=None,scale=False, test_size = None, n_repeats = 30,**kwargs):  
        """
        Fit the model to input data X and target values Y, with extra optional 
        arguments for preprocessing data.
        
        Parameters
        ----------
        x : ndarray or sparse matrix of shape (n_samples, n_features)
            Input data.
        y : ndarray of shape (n_samples,)
            Target values.
        decor : tuple or float, optional
            Decorrellation parameter. If a number in range [0,1] is chosen
            then the input features that have a correllation factor above the 
            give value, are removed. If tuple of (string, float [0,1]) is chosen
            then the input data features are ranked based on their feature
            importance method, indicated by the string, then decorrelated and the
            ones with higher importance are retained. The string could be one
            from the following:                
                "variance"
                "f_test"
                "mutual_info"
                "chi2"
                "pearson"                
            The default is None.
        scale : bool, optional
            Wether to scale the input data or not (centering and scaling).
            The default is False.
        test_size : float, optional
            Represents the portion of input data to be included for test split. 
            Needs to be between 0 and 1. The default is None.
        n_repeats : int, optional
            Number of times to permute the feature for the calculation of 
            permutation importance. The default is 30.
            
        Attributes
        ----------
        best_fits: Vector of shape (n_iter,)
            Fitness values of each iteration.        
        best_sol: Vector of shape (n_f,)
            Indices of selected fetures after preprocessing input data.
        best_sol_acc: float
            Optimized fitness value for the identified solution.
        model_best: Class
            Regression/classification estimator given to the optimization 
            algorithm. At this stage model_best = model
        rankings: rankings vectors from the method proposed by Hassanzadeh et al. (2021)            

        Returns
        -------
        Vector of shape (n_f,)
            Indices of selected fetures of the original input data (prior to 
            preprocessing). If no deocr is given, this equals to best_sol.
        Vector of shape (n_iter,)
            Fitness values of each iteration.

        """
        self.load(x,y,decor=decor,scale=scale)        
        if test_size is not None:
            self.train_test_split(test_size, random_state=self.random_state,**kwargs)
                
        a = time.time()
        if hasattr(self, 'x_train') and hasattr(self, 'x_test') and hasattr(self, 'y_train') and hasattr(self, 'y_test'):
            self.alg_opt = _AntColony(self.model,
                                      self.cv,
                                      self.scorer,
                                      self.n_f,
                                      self.fitness,
                                      self.weight,
                                      self.x_train,
                                      self.y_train,                                                                             
                                      x_test = self.x_test,
                                      y_test = self.y_test,
                                      MaxIt = self.n_iter, 
                                      n_ant = self.n_ant,
                                      Q = self.Q,
                                      tau0 = self.tau0,
                                      alpha = self.alpha,
                                      beta = self.beta,
                                      rho = self.rho,
                                      verbose = self.verbose,
                                      random_state = self.random_state)     
            
        else:
            self.alg_opt = _AntColony(self.model,
                                      self.cv,
                                      self.scorer,
                                      self.n_f,
                                      self.fitness,
                                      self.weight,
                                      self.x,
                                      self.y,
                                      MaxIt = self.n_iter, 
                                      n_ant = self.n_ant,
                                      Q = self.Q,
                                      tau0 = self.tau0,
                                      alpha = self.alpha,
                                      beta = self.beta,
                                      rho = self.rho,
                                      verbose = self.verbose,
                                      random_state = self.random_state)     


        best_sol, self.best_fits, tops = self.alg_opt.generate()
        self.best_sol = best_sol[0]
        self.best_sol_acc = best_sol[1]       
        self.best_features = self.best_sol
        # setting the model_best = model since one may not want to run tune_param
        self.model_best = clone(self.model)
        
        #setting top rankings 
        self.rankings = self._feature_ranking(tops, self.x_cols, n_repeats=n_repeats, random_state= self.random_state)
        
        if self.verbose:
            print("Optimization completed in {} seconds".format(time.time() - a))
            print("Best individual is {}, {:.4}".format(np.array(self.x_cols[self.best_sol]), self.best_sol_acc))
        
        return self.x_cols[self.best_sol], self.best_fits  
    
    @property
    def _name_(self):
        return "ACO"
        
        
    
    

        
        
        


import numpy as np
from ..metrics._metrics import cross_validate_LOO, cross_val_predict_loo
from sklearn.metrics import accuracy_score
from sklearn.metrics import mean_squared_error
from sklearn.metrics import r2_score
from sklearn.base import is_regressor
from sklearn.model_selection import cross_val_predict
import warnings
from sklearn.metrics import make_scorer

def calculate_fitness(model,cv,scorer, x,y,x_test = None, y_test = None, n_jobs = -1):
    '''
    This function calculates fitness of the function (loss/accuracy measures).
    '''
    
    n_features = x.shape[1] 
    if cv is not None:
        if (x_test is not None) and (y_test is not None):
            raise ValueError("Either test partition (x_test and y_test) or cv should be defined.")            
        if scorer.__name__ == "adj_r2_score":
            raise ModuleNotFoundError("adj_r2_score is not availale with CV.")                 
        
        # if the model is a regressor
        if is_regressor(model):                                
            if cv.__str__() == "LeaveOneOut()":
                y_pred = cross_val_predict_loo(model,x,y,cv)                    
            else:
                y_pred = cross_val_predict(model,x,y, n_jobs = n_jobs, cv=cv)
            score = scorer(y,y_pred)                            
                              
        # if the model is a classifier                
        else:
            if cv.__str__() == "LeaveOneOut()":
                y_pred = cross_val_predict_loo(model,x,y,cv)                       
                                            
            else:
                y_pred = cross_val_predict(model,x,y,n_jobs = n_jobs, cv=cv)     
            score = scorer(y,y_pred)            
        
        
    elif cv is None:        
        # if regression
        if is_regressor(model):    
            if (x_test is not None) and (y_test is not None):
                model.fit(x,y)
                y_pred = model.predict(x_test)                  
            else:
                #warnings.warn("Neither cv_model nor test partition (x_test/y_test) are defined, predicting on x/y.")
                model.fit(x,y)
                y_pred = model.predict(x)   
                y_test = y 
            score = scorer(y_test,y_pred)    
            
        # if classification                    
        else:
            if (x_test is not None) and (y_test is not None):
                model.fit(x,y)
                y_pred = model.predict(x_test)                   
            else:
                #warnings.warn("Neither cv_model nor test partition (x_test/y_test) are defined, predicting on x/y.")
                model.fit(x,y)
                y_pred = model.predict(x)       
                y_test = y
            score = scorer(y_test,y_pred)                          
             
    return (score, n_features)            


# -*- coding: utf-8 -*-
"""
Created on Tue Sep 22 09:27:57 2020

@author: Amirh
"""
from ..metrics._metrics import adj_r2_score, kappa_score, rmse_score, rmspe_score
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import warnings
import time
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import r2_score
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import RandomizedSearchCV
from sklearn.metrics import roc_curve, auc
from sklearn.metrics import confusion_matrix
from sklearn.base import is_regressor, is_classifier
from sklearn.model_selection import train_test_split
from sklearn.model_selection import cross_validate, cross_val_predict
from sklearn.metrics import *
from sklearn.feature_selection import VarianceThreshold, f_regression, f_classif, \
    mutual_info_regression, mutual_info_classif, chi2
from scipy import stats
from sklearn.utils import check_X_y, check_array
from sklearn.utils.multiclass import type_of_target, unique_labels
from ..metrics._metrics import cross_validate_LOO
from scipy.optimize import minimize
from sklearn.inspection import permutation_importance
from copy import copy
from functools import partial, update_wrapper

class BaseFeatureSelector(object):
    def __init__(self,n_f=None,scoring = None,verbose = False,**kwargs):
        self.n_f = n_f
        self.scoring = scoring
        # check scoring
        #self.scorer = self._check_scoring()   
        self.scorer = self._scorer_maker(**kwargs)        
        self.verbose = verbose
    
    def load(self,x,y,decor=None,scale=False):
        """
        This function loads the data, decorrellates and scales the given data.
        """
        
        # load data    
        x = check_array(x)           
        x,y = check_X_y(x,y)
        self.x = x
                
        # checking to see if the model is regression or classification
        if is_regressor(self.model) and (type_of_target(y) == "continuous" or 
                                         type_of_target(y) == "multiclass"): #for floats that are whole numbers
            self.y = y
        elif (is_classifier(self.model)) and (type_of_target(y) == "multiclass" or 
                                              type_of_target(y) == "binary"):
            self.y = y
            self.class_sep = unique_labels(self.y).astype(np.int16)   

        else:
            raise ValueError("Mismatch between objective model and the input data.")                     
        if self.verbose:            
            print("Data successfully loaded.")  
        
        # scale data
        if scale:
            scaler = StandardScaler()
            self.x = scaler.fit_transform(self.x)
            if self.verbose:
                print("Data standardized")    
                
        # decorrellate data
        if decor is not None:
            x_df = pd.DataFrame(self.x)
            if type(decor) is tuple:
                indices = self._feature_importance(decor[0])                
                x_df = x_df[indices] # sorting features
                decor = decor[1]
                
            if decor == int(decor):
                x_df = self._decorrellate_n(x_df, decor)
                self.x_cols = x_df.columns    
                self.x = x_df.values
            
            if decor == float(decor):
                x_df = self._decorrellate_thr(x_df, decor)
                self.x_cols = x_df.columns    
                self.x = x_df.values
                
            
            else:
	            raise ValueError('Decorrellation value should be either int for number of features or float for pearson decorrellation.')


            if self.verbose:
                print("Data decorrellated to {} features.".format(self.x.shape[1]))
        else:
        	self.x_cols = pd.DataFrame(self.x).columns    

            
    def train_test_split(self, test_size, random_state=1,**kwargs):
        """
        The function splits the data into train and test size.        
        """
        
        if self.cv:
            raise ValueError("Either train/test partition or cv should be defined.")     
        if test_size > 0 and test_size < 1:
            self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(self.x, self.y, test_size=test_size, random_state=random_state,**kwargs)
        else:
            raise ValueError('test_size should be bigger than 0 and smaller than 1.') 
        if self.verbose:            
            print("train/test split created with test size of {}.".format(test_size))            

    
    def tune_model_params(self,param_dict,random_state = 1):
        """
        Function tuning model parameters using RandomizedCVSearch.

        Parameters
        ----------
        param_dict : dictionary
            Dictionary containing parameters of the given model.
        random_state : int, optional
            Seed for random state. The default is 1.

        Returns
        -------
        Estimator
            Function returns the best estimator chosen by RandomizedCVSearch.

        """        
        a = time.time()
        Searcher = RandomizedSearchCV(self.model, param_dict, random_state=random_state,n_jobs=-1)
        if hasattr(self, 'x_train') and hasattr(self, 'x_test') and hasattr(self, 'y_train') and hasattr(self, 'y_test'):
            search = Searcher.fit(self.x_train[:,self.best_features],self.y_train)
        else:
            search = Searcher.fit(self.x[:,self.best_features],self.y)
        if self.verbose:
            print("Parameter tuning completed in {:.2f} seconds".format(time.time() - a))            
            print("best params are: \n",search.best_params_)
        self.model_best = search.best_estimator_
        
        return search.best_estimator_

    def display_results(self):
        """
        Function displaying results for the best identified set of features.

        Returns
        -------
        Accuracy metrics for the best identified set of features.

        """
        
        if self.cv is not None:
            # if task is regression
            if is_regressor(self.model):
# =============================================================================
#           Running with general algorithm and calculating accuracies
# =============================================================================                

                # calculating accuracies for generalized model             
                self.model.fit(self.x,self.y)
                self.y_pred_cal_gen = np.squeeze(self.model.predict(self.x)).squeeze()
                self.r2_cal_gen = r2_score(self.y,self.y_pred_cal_gen)
                self.r2_adj_cal_gen = adj_r2_score(self.y,self.y_pred_cal_gen, self.x, self.r2_cal_gen)
                self.rmse_cal_gen = mean_squared_error(self.y, self.y_pred_cal_gen, squared = False)
                self.rmspe_cal_gen = rmspe_score(self.y, self.y_pred_cal_gen)

                if self.cv.__str__() == "LeaveOneOut()":
                    score_r2 = cross_validate_LOO(self.model,self.x,self.y,self.cv,
                                                  scoring=r2_score)                    

                    score_rmse = cross_validate_LOO(self.model,self.x,self.y,self.cv,
                                                    scoring=mean_squared_error)
                    
                    self.y_pred_cv_gen = cross_val_predict(self.model, self.x, self.y, n_jobs = -1, cv=self.cv).squeeze()        
                    self.r2_cv_gen = score_r2          
                    self.r2_adj_cv_gen = adj_r2_score(self.y, x=self.x, r2=self.r2_cv_gen)
                    self.rmse_cv_gen = np.sqrt(score_rmse)
                    self.rmspe_cv_gen = rmspe_score(self.y, self.y_pred_cv_gen)                    
                    
                else:
                    scores = cross_validate(self.model,self.x,self.y,
                                                  n_jobs = -1, 
                                                  cv=self.cv,
                                                  scoring=['r2','neg_mean_squared_error'])    
                    
                    self.y_pred_cv_gen = cross_val_predict(self.model, self.x, self.y, n_jobs = -1, cv=self.cv).squeeze()        
                    self.r2_cv_gen = scores['test_r2'].mean()            
                    self.r2_adj_cv_gen = adj_r2_score(self.y, x=self.x, r2=self.r2_cv_gen)
                    self.rmse_cv_gen = np.sqrt(scores['test_neg_mean_squared_error'].mean())
                    self.rmspe_cv_gen = rmspe_score(self.y, self.y_pred_cv_gen)                                             

                
    # =============================================================================
    #           Running optimized model and calculate accuracies
    # =============================================================================
                    
                self.model_best.fit(self.x[:,self.best_features],self.y)            
                self.y_pred_cal_op = np.squeeze(self.model_best.predict(self.x[:,self.best_features])).squeeze()
                self.r2_cal_op = r2_score(self.y,self.y_pred_cal_op)
                self.r2_adj_cal_op = adj_r2_score(self.y,self.y_pred_cal_op,self.x[:,self.best_features],self.r2_cal_op)
                self.rmse_cal_op = mean_squared_error(self.y, self.y_pred_cal_op, squared = False)
                self.rmspe_cal_op = rmspe_score(self.y, self.y_pred_cal_op)
                                
                if self.cv.__str__() == "LeaveOneOut()":
                    
                    score_r2 = cross_validate_LOO(self.model_best,self.x[:,self.best_features],self.y,self.cv,
                                                  scoring=r2_score)                    

                    score_rmse = cross_validate_LOO(self.model_best,self.x[:,self.best_features],self.y,self.cv,
                                                    scoring=mean_squared_error)                    
                    
                    self.y_pred_cv_op = cross_val_predict(self.model_best, self.x[:,self.best_features], self.y, n_jobs = -1, cv=self.cv).squeeze()                                    
                    self.r2_cv_op = score_r2   
                    self.r2_adj_cv_op = adj_r2_score(self.y,x = self.x[:,self.best_features],r2= self.r2_cv_op)
                    self.rmse_cv_op = np.sqrt(score_rmse)
                    self.rmspe_cv_op = rmspe_score(self.y, self.y_pred_cv_op)                    
                    

                else:
                    scores = cross_validate(self.model_best,self.x[:,self.best_features],self.y,
                                            n_jobs = -1, 
                                            cv=self.cv,
                                            scoring=['r2','neg_mean_squared_error'])                    
                    
                    self.y_pred_cv_op = cross_val_predict(self.model_best, self.x[:,self.best_features], self.y, n_jobs = -1, cv=self.cv).squeeze()                                    
                    self.r2_cv_op = scores['test_r2'].mean()     
                    self.r2_adj_cv_op = adj_r2_score(self.y,x = self.x[:,self.best_features],r2= self.r2_cv_op)
                    self.rmse_cv_op = np.sqrt(scores['test_neg_mean_squared_error'].mean())
                    self.rmspe_cv_op = rmspe_score(self.y, self.y_pred_cv_op)

    # =============================================================================
    #           Plotting generalized model and optimized model        
    # =============================================================================
                # generalized model
                fig,axes = plt.subplots(2,3,figsize=(15,10))
                sns.regplot(x=self.y, y=self.y_pred_cal_gen, color="k",ax=axes[0, 0])
                axes[0,0].plot([np.min([self.y, self.y_pred_cal_gen]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cal_gen]) + 0.1*np.std(self.y)],
                    [np.min([self.y, self.y_pred_cal_gen]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cal_gen]) + 0.1*np.std(self.y)],'r--')
                axes[0,0].text(0.05,0.80,r"RMSE={:.3f}".format(self.rmse_cal_gen) + "\n"+ 
                                       r"RMSPE={:.3f}".format(self.rmspe_cal_gen) + "\n"+ 
                                       r"$R^2$={:.3f}".format(self.r2_cal_gen) + "\n"+
                                       r"$Adj. R^2$={:.3f}".format(self.r2_adj_cal_gen),
                                       transform=axes[0, 0].transAxes,
                                       color='k',
                                       bbox=dict(facecolor=(1,1,1,0.5),edgecolor=(0,0,0,1)))
                axes[0,0].set_xlabel("Measured")
                axes[0,0].set_ylabel("Predicted")
                axes[0,0].set_title("General model - Cal.")
                axes[0,0].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))
                
                sns.regplot(x=self.y, y=self.y_pred_cv_gen, color="k",ax=axes[1, 0])
                axes[1,0].plot([np.min([self.y, self.y_pred_cv_gen]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cv_gen]) + 0.1*np.std(self.y)],
                    [np.min([self.y, self.y_pred_cv_gen]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cv_gen]) + 0.1*np.std(self.y)],'r--')            
                axes[1,0].text(0.05,0.80,r"RMSE={:.3f}".format(self.rmse_cv_gen) + "\n"+ 
                                       r"RMSPE={:.3f}".format(self.rmspe_cv_gen) + "\n"+ 
                                       r"$R^2$={:.3f}".format(self.r2_cv_gen) + "\n"+
                                       r"$Adj. R^2$={:.3f}".format(self.r2_adj_cv_gen),
                                       transform=axes[1, 0].transAxes,
                                       color='k',
                                       bbox=dict(facecolor=(1,1,1,0.5),edgecolor=(0,0,0,1)))
                axes[1,0].set_xlabel("Measured")
                axes[1,0].set_ylabel("Predicted")
                axes[1,0].set_title("General model - CV")
                axes[1,0].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))
    
                # optimized model
                sns.regplot(x=self.y, y=self.y_pred_cal_op, color="k",ax=axes[0, 1])
                axes[0,1].plot([np.min([self.y, self.y_pred_cal_op]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cal_op]) + 0.1*np.std(self.y)],
                    [np.min([self.y, self.y_pred_cal_op]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cal_op]) + 0.1*np.std(self.y)],'r--')               
                axes[0,1].text(0.05,0.80,r"RMSE={:.3f}".format(self.rmse_cal_op) + "\n"+ 
                                       r"RMSPE={:.3f}".format(self.rmspe_cal_op) + "\n"+ 
                                       r"$R^2$={:.3f}".format(self.r2_cal_op)+ "\n"+
                                       r"$Adj. R^2$={:.3f}".format(self.r2_adj_cal_op),
                                       transform=axes[0, 1].transAxes,
                                       color='k',
                                       bbox=dict(facecolor=(1,1,1,0.5),edgecolor=(0,0,0,1)))
                axes[0,1].set_xlabel("Measured")
                axes[0,1].set_ylabel("Predicted")
                axes[0,1].set_title("Optimized model - Cal.")
                axes[0,1].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                   
                
                sns.regplot(x=self.y, y=self.y_pred_cv_op, color="k",ax=axes[1, 1])
                axes[1,1].plot([np.min([self.y, self.y_pred_cv_op]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cv_op]) + 0.1*np.std(self.y)],
                    [np.min([self.y, self.y_pred_cv_op]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cv_op]) + 0.1*np.std(self.y)],'r--')              
                axes[1,1].text(0.05,0.80,r"RMSE={:.3f}".format(self.rmse_cv_op) + "\n"+ 
                                       r"RMSPE={:.3f}".format(self.rmspe_cv_op) + "\n"+ 
                                       r"$R^2$={:.3f}".format(self.r2_cv_op)+ "\n"+
                                       r"$Adj. R^2$={:.3f}".format(self.r2_adj_cv_op),
                                       transform=axes[1, 1].transAxes,
                                       color='k',
                                       bbox=dict(facecolor=(1,1,1,0.5),edgecolor=(0,0,0,1)))
                axes[1,1].set_xlabel("Measured")
                axes[1,1].set_ylabel("Predicted")
                axes[1,1].set_title("Optimized model - CV")
                axes[1,1].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                                   
                
                #residuals plot
                sns.residplot(self.y_pred_cal_op, self.y - self.y_pred_cal_op, lowess=False, color="k",ax=axes[0, 2])
                axes[0,2].set_xlabel("Predicted")
                axes[0,2].set_ylabel("Residuals")
                axes[0,2].set_title("Optimized model - Cal. Residuals")
                axes[0,2].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                   

                
                sns.residplot(self.y_pred_cv_op, self.y - self.y_pred_cv_op, lowess=False, color="k",ax=axes[1, 2])
                axes[1,2].set_xlabel("Predicted")
                axes[1,2].set_ylabel("Residuals")
                axes[1,2].set_title("Optimized model - CV Residuals")
                axes[1,2].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                      
                
                

                results = {"r2_cal_gen":[self.r2_cal_gen],'rmse_cal_gen':[self.rmse_cal_gen],'rmspe_cal_gen':[self.rmspe_cal_gen],'r2_adj_cal_gen':[self.r2_adj_cal_gen],
                           'r2_cv_gen':[self.r2_cv_gen],'rmse_cv_gen':[self.rmse_cv_gen],'rmspe_cv_gen':[self.rmspe_cv_gen],'r2_adj_cv_gen':[self.r2_adj_cv_gen],
                           "r2_cal_op":[self.r2_cal_op],'rmse_cal_op':[self.rmse_cal_op],'rmspe_cal_op':[self.rmspe_cal_op],'r2_adj_cal_op':[self.r2_adj_cal_op],
                           'r2_cv_op':[self.r2_cv_op],'rmse_cv_op':[self.rmse_cv_op],'rmspe_cv_op':[self.rmspe_cv_op],'r2_adj_cv_op':[self.r2_adj_cv_op]}    
                    
            # if task is classification
            else:
                if len(self.class_sep) == 2:
                    fig,axes = plt.subplots(2,3,figsize=(15,10))
                else:
                    fig,axes = plt.subplots(2,2,figsize=(10 + len(self.class_sep),10 + len(self.class_sep)))
                # general model
                
                self.model.fit(self.x,self.y)
                y_train_pred  = self.model.predict(self.x)
                self.acc_train_gen = self.model.score(self.x,self.y)
                cm = confusion_matrix(self.y,y_train_pred.squeeze())
                cm = np.stack(list(map(lambda x: x/x.sum(),cm)))
                conf_matrix=pd.DataFrame(cm,columns = self.class_sep,index= self.class_sep)
                sns.heatmap(conf_matrix,cmap="coolwarm", annot=True, ax=axes[0,0],vmin=0,vmax=1)
                axes[0,0].set_title('General model - Cal. confusion matrix')
                axes[0,0].set_ylabel('True label')
                axes[0,0].set_xlabel('Predicted label')            
                                
                
                y_test_pred  = cross_val_predict(self.model, self.x, self.y, 
                             n_jobs = -1,
                             cv=self.cv).squeeze()                  

                if self.cv.__str__() == "LeaveOneOut()":          
                    score_acc = cross_validate_LOO(self.model,self.x,self.y,self.cv,
                                                  scoring=accuracy_score)                                                    

                    self.acc_test_gen = score_acc           
                else:
                    scores = cross_validate(self.model, self.x, self.y, 
                                            n_jobs = -1,
                                            cv=self.cv,
                                            scoring = "accuracy")                    
                    
                    self.acc_test_gen = scores['test_score'].mean()            

                cm = confusion_matrix(self.y,y_test_pred)
                cm = np.stack(list(map(lambda x: x/x.sum(),cm)))
                conf_matrix=pd.DataFrame(cm,columns = self.class_sep,index= self.class_sep)
                sns.heatmap(conf_matrix,cmap="coolwarm", annot=True,ax = axes[1,0],vmin=0,vmax=1)
                axes[1,0].set_title('General model - CV confusion matrix')  
                axes[1,0].set_ylabel('True label')
                axes[1,0].set_xlabel('Predicted label')           
    
                
                # optimized model
                self.model_best.fit(self.x[:,self.best_features],self.y)
                y_train_pred  = self.model_best.predict(self.x[:,self.best_features])
                self.acc_train_op = self.model_best.score(self.x[:,self.best_features],self.y)
                cm = confusion_matrix(self.y, y_train_pred.squeeze())
                cm = np.stack(list(map(lambda x: x/x.sum(),cm)))
                conf_matrix=pd.DataFrame(cm,columns = self.class_sep,index= self.class_sep)
                sns.heatmap(conf_matrix,cmap="coolwarm", annot=True, ax=axes[0,1],vmin=0,vmax=1)
                axes[0,1].set_title('Optimized model - Cal. confusion matrix')     
                axes[0,1].set_ylabel('True label')
                axes[0,1].set_xlabel('Predicted label')        
                        
                    
                y_test_pred  = cross_val_predict(self.model_best, self.x[:,self.best_features], self.y, 
                                                 n_jobs = -1,
                                                 cv=self.cv).squeeze()  
                
                if self.cv.__str__() == "LeaveOneOut()":   
                    score_acc = cross_validate_LOO(self.model_best,self.x[:,self.best_features],self.y,
                                                   self.cv,
                                                   scoring=accuracy_score)                        
                    
                    self.acc_test_op = score_acc                                 
                else:
                    scores = cross_validate(self.model_best, self.x[:,self.best_features], self.y, 
                                            n_jobs = -1,
                                            cv=self.cv,
                                            scoring = "accuracy")                              

                    self.acc_test_op = scores['test_score'].mean()
                    
                cm = confusion_matrix(self.y,y_test_pred)
                cm = np.stack(list(map(lambda x: x/x.sum(),cm)))
                conf_matrix=pd.DataFrame(cm,columns = self.class_sep,index= self.class_sep)
                sns.heatmap(conf_matrix,cmap="coolwarm", annot=True,ax = axes[1,1],vmin=0,vmax=1)
                axes[1,1].set_title('Optimized model - CV confusion matrix')      
                axes[1,1].set_ylabel('True label')
                axes[1,1].set_xlabel('Predicted label')    
                
                if len(self.class_sep) == 2:
                    #optimized model Roc curves
                    self.model_best.fit(self.x[:,self.best_features],self.y)
                    y_train_pred_prob = self.model_best.predict_proba(self.x[:,self.best_features])
                    fpr, tpr, thr = roc_curve(self.y, y_train_pred_prob[:, 1].squeeze())   
                    self.auc_train_op = auc(fpr, tpr)
                    axes[0,2].plot([0,1],[0,1],linestyle = '--',lw = 2,color = 'black')
                    axes[0,2].plot(fpr, tpr, color='blue',
                             label=r'$ROC_{train}$ (AUC = %0.2f)' % (auc(fpr, tpr)),lw=2, alpha=1)
                    axes[0,2].set_xlabel('False Positive Rate')
                    axes[0,2].set_ylabel('True Positive Rate')
                    axes[0,2].set_title('Optimized model - Cal. ROC Curve')
                    axes[0,2].legend(loc="lower right")
                    
                    y_test_pred_prob  = cross_val_predict(self.model_best, self.x[:,self.best_features], self.y, 
                                                     n_jobs = -1,
                                                     cv=self.cv,
                                                     method='predict_proba')    
                
                    
                    fpr, tpr, thr = roc_curve(self.y, y_test_pred_prob[:, 1].squeeze())      
                    self.auc_test_op = auc(fpr, tpr)
                    axes[1,2].plot([0,1],[0,1],linestyle = '--',lw = 2,color = 'black')
                    axes[1,2].plot(fpr, tpr, color='blue',
                             label=r'$ROC_{test}$ (AUC = %0.2f)' % (auc(fpr, tpr)),lw=2, alpha=1)
                    axes[1,2].set_xlabel('False Positive Rate')
                    axes[1,2].set_ylabel('True Positive Rate')
                    axes[1,2].set_title('Optimized model- CV ROC Curve')
                    axes[1,2].legend(loc="lower right")
                
                    results = {"acc_gen":[self.acc_train_gen],'acc_cv_gen':[self.acc_test_gen],
                       'acc_op':[self.acc_train_op],'acc_cv_op':[self.acc_test_op],
                       "auc_op":[self.auc_train_op],'auc_cv_op':[self.auc_test_op]}
                
                results = {"acc_gen":[self.acc_train_gen],'acc_cv_gen':[self.acc_test_gen],
                   'acc_op':[self.acc_train_op],'acc_cv_op':[self.acc_test_op]}                                
            
        # if cv is not defined
        else:
            if is_regressor(self.model):
                if hasattr(self, 'x_train') and hasattr(self, 'x_test') and hasattr(self, 'y_train') and hasattr(self, 'y_test'):         

                    # calculating accuracies for generalized model
                    self.model.fit(self.x_train,self.y_train)
                    self.y_pred_cal_gen = np.squeeze(self.model.predict(self.x_train)).squeeze()
                    self.r2_cal_gen = r2_score(self.y_train,self.y_pred_cal_gen)
                    #self.r2_adj_cal_gen = adj_r2_score(self.y_train,self.y_pred_cal_gen,x = self.x_train, r2= self.r2_cal_gen)
                    self.r2_adj_cal_gen = adj_r2_score(self.y_train,self.y_pred_cal_gen,x = self.x, r2= self.r2_cal_gen)
                    self.rmse_cal_gen = mean_squared_error(self.y_train, self.y_pred_cal_gen, squared = False)
                    self.rmspe_cal_gen = rmspe_score(self.y_train, self.y_pred_cal_gen)
                    
                    self.y_pred_val_gen = np.squeeze(self.model.predict(self.x_test)).squeeze()
                    self.r2_val_gen = r2_score(self.y_test,self.y_pred_val_gen)
                    #self.r2_adj_val_gen = adj_r2_score(self.y_test, x=self.x_test, r2=self.r2_val_gen)
                    self.r2_adj_val_gen = adj_r2_score(self.y_test, x=self.x, r2=self.r2_val_gen)
                    self.rmse_val_gen = mean_squared_error(self.y_test, self.y_pred_val_gen, squared = False)
                    self.rmspe_val_gen = rmspe_score(self.y_test, self.y_pred_val_gen)
                    
                    # calculating accuracies for optimized model
                    self.model_best.fit(self.x_train[:,self.best_features],self.y_train)
                    self.y_pred_cal_op = np.squeeze(self.model_best.predict(self.x_train[:,self.best_features])).squeeze()
                    self.r2_cal_op = r2_score(self.y_train,self.y_pred_cal_op)
                    #self.r2_adj_cal_op = adj_r2_score(self.y_train, x=self.x_train[:,self.best_features],r2=self.r2_cal_op)
                    self.r2_adj_cal_op = adj_r2_score(self.y_train, x=self.x[:,self.best_features],r2=self.r2_cal_op)
                    self.rmse_cal_op = mean_squared_error(self.y_train, self.y_pred_cal_op, squared = False)  
                    self.rmspe_cal_op = rmspe_score(self.y_train, self.y_pred_cal_op)
                    
                    self.y_pred_val_op = np.squeeze(self.model_best.predict(self.x_test[:,self.best_features])).squeeze()
                    self.r2_val_op = r2_score(self.y_test,self.y_pred_val_op)
                    #self.r2_adj_val_op = adj_r2_score(self.y_test, x= self.x_test[:,self.best_features],r2=self.r2_val_op)
                    self.r2_adj_val_op = adj_r2_score(self.y_test, x= self.x[:,self.best_features],r2=self.r2_val_op)
                    self.rmse_val_op = mean_squared_error(self.y_test, self.y_pred_val_op, squared = False)
                    self.rmspe_val_op = rmspe_score(self.y_test, self.y_pred_val_op)
                    
                    # generalized model
                    fig,axes = plt.subplots(2,3,figsize=(15,10))
                    sns.regplot(x=self.y_train, y=self.y_pred_cal_gen, color="k",ax=axes[0, 0])
                    axes[0,0].plot([np.min([self.y_train, self.y_pred_cal_gen]) - 0.1*np.std(self.y), np.max([self.y_train, self.y_pred_cal_gen]) + 0.1*np.std(self.y)],
                        [np.min([self.y_train, self.y_pred_cal_gen]) - 0.1*np.std(self.y), np.max([self.y_train, self.y_pred_cal_gen]) + 0.1*np.std(self.y)],'r--')
                    axes[0,0].text(0.05,0.80,r"RMSE={:.3f}".format(self.rmse_cal_gen) + "\n"+ 
                                       r"RMSPE={:.3f}".format(self.rmspe_cal_gen) + "\n"+ 
                                      r"$R^2$={:.3f}".format(self.r2_cal_gen)+ "\n"+
                                      r"$Adj. R^2$={:.3f}".format(self.r2_adj_cal_gen),
                                      transform=axes[0, 0].transAxes,
                                      color='k',
                                      bbox=dict(facecolor=(1,1,1,0.5),edgecolor=(0,0,0,1)))
                    axes[0,0].set_xlabel("Measured")
                    axes[0,0].set_ylabel("Predicted")
                    axes[0,0].set_title("General model - Cal.")
                    axes[0,0].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                                       
                    
                    sns.regplot(x=self.y_test, y=self.y_pred_val_gen, color="k",ax=axes[1, 0])
                    axes[1,0].plot([np.min([self.y_test, self.y_pred_val_gen]) - 0.1*np.std(self.y), np.max([self.y_test, self.y_pred_val_gen]) + 0.1*np.std(self.y)],
                        [np.min([self.y_test, self.y_pred_val_gen]) - 0.1*np.std(self.y), np.max([self.y_test, self.y_pred_val_gen]) + 0.1*np.std(self.y)],'r--')            
                    axes[1,0].text(0.05,0.80,r"RMSE={:.3f}".format(self.rmse_val_gen) + "\n"+ 
                                       r"RMSPE={:.3f}".format(self.rmspe_val_gen) + "\n"+ 
                                      r"$R^2$={:.3f}".format(self.r2_val_gen)+ "\n"+
                                      r"$Adj. R^2$={:.3f}".format(self.r2_adj_val_gen),
                                      transform=axes[1, 0].transAxes,
                                      color='k',
                                      bbox=dict(facecolor=(1,1,1,0.5),edgecolor=(0,0,0,1)))
                    axes[1,0].set_xlabel("Measured")
                    axes[1,0].set_ylabel("Predicted")
                    axes[1,0].set_title("General model - Val.")
                    axes[1,0].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                               

                    # optimized model
                    sns.regplot(x=self.y_train, y=self.y_pred_cal_op, color="k",ax=axes[0, 1])
                    axes[0,1].plot([np.min([self.y_train, self.y_pred_cal_op]) - 0.1*np.std(self.y), np.max([self.y_train, self.y_pred_cal_op]) + 0.1*np.std(self.y)],
                        [np.min([self.y_train, self.y_pred_cal_op]) - 0.1*np.std(self.y), np.max([self.y_train, self.y_pred_cal_op]) + 0.1*np.std(self.y)],'r--')               
                    axes[0,1].text(0.05,0.80,r"RMSE={:.3f}".format(self.rmse_cal_op) + "\n"+ 
                                       r"RMSPE={:.3f}".format(self.rmspe_cal_op) + "\n"+ 
                                      r"$R^2$={:.3f}".format(self.r2_cal_op)+ "\n"+
                                      r"$Adj. R^2$={:.3f}".format(self.r2_adj_cal_op),
                                      transform=axes[0, 1].transAxes,
                                      color='k',
                                      bbox=dict(facecolor=(1,1,1,0.5),edgecolor=(0,0,0,1)))
                    axes[0,1].set_xlabel("Measured")
                    axes[0,1].set_ylabel("Predicted")
                    axes[0,1].set_title("Optimized model - Cal.")
                    axes[0,1].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                                                         
                    
                    sns.regplot(x=self.y_test, y=self.y_pred_val_op, color="k",ax=axes[1, 1])
                    axes[1,1].plot([np.min([self.y_test, self.y_pred_val_op]) - 0.1*np.std(self.y), np.max([self.y_test, self.y_pred_val_op]) + 0.1*np.std(self.y)],
                        [np.min([self.y_test, self.y_pred_val_op]) - 0.1*np.std(self.y), np.max([self.y_test, self.y_pred_val_op]) + 0.1*np.std(self.y)],'r--')              
                    axes[1,1].text(0.05,0.80,r"RMSE={:.3f}".format(self.rmse_val_op) + "\n"+ 
                                      r"RMSPE={:.3f}".format(self.rmspe_val_op) + "\n"+ 
                                      r"$R^2$={:.3f}".format(self.r2_val_op)+ "\n"+
                                      r"$Adj. R^2$={:.3f}".format(self.r2_adj_val_op),
                                      transform=axes[1, 1].transAxes,
                                      color='k',
                                      bbox=dict(facecolor=(1,1,1,0.5),edgecolor=(0,0,0,1)))
                    axes[1,1].set_xlabel("Measured")
                    axes[1,1].set_ylabel("Predicted")
                    axes[1,1].set_title("Optimized model - Val.")
                    axes[1,1].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                               
                    
                    #residuals plot
                    sns.residplot(self.y_pred_cal_op, self.y_train - self.y_pred_cal_op, lowess=False, color="k",ax=axes[0, 2])
                    axes[0,2].set_xlabel("Predicted")
                    axes[0,2].set_ylabel("Residuals")
                    axes[0,2].set_title("Optimized model - Cal. Residuals")
                    axes[0,2].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                     
                    
                    
                    sns.residplot(self.y_pred_val_op, self.y_test - self.y_pred_val_op, lowess=False, color="k",ax=axes[1, 2])
                    axes[1,2].set_xlabel("Predicted")
                    axes[1,2].set_ylabel("Residuals")
                    axes[1,2].set_title("Optimized model - Val. Residuals")        
                    axes[1,2].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                                       
                    
                    results = {"r2_cal_gen":[self.r2_cal_gen],'rmse_cal_gen':[self.rmse_cal_gen], 'rmspe_cal_gen':[self.rmspe_cal_gen], 'r2_adj_cal_gen':[self.r2_adj_cal_gen],
                               'r2_val_gen':[self.r2_val_gen],'rmse_val_gen':[self.rmse_val_gen],'rmspe_val_gen':[self.rmspe_val_gen], 'r2_adj_val_gen':[self.r2_adj_val_gen],
                               "r2_cal_op":[self.r2_cal_op],'rmse_cal_op':[self.rmse_cal_op],'rmspe_cal_op':[self.rmspe_cal_op], 'r2_adj_cal_op':[self.r2_adj_cal_op],
                               'r2_val_op':[self.r2_val_op],'rmse_val_op':[self.rmse_val_op],'rmspe_val_op':[self.rmspe_val_op], 'r2_adj_val_op':[self.r2_adj_val_op]}      
                    
                # test partitions not defined 
                else:
                    # calculating accuracies for generalized model
                    self.model.fit(self.x,self.y)
                    self.y_pred_cal_gen = np.squeeze(self.model.predict(self.x)).squeeze()
                    self.r2_cal_gen = r2_score(self.y,self.y_pred_cal_gen)
                    self.r2_adj_cal_gen = adj_r2_score(self.y,x=self.x, r2=self.r2_cal_gen)
                    self.rmse_cal_gen = mean_squared_error(self.y, self.y_pred_cal_gen, squared = False)    
                    self.rmspe_cal_gen = rmspe_score(self.y, self.y_pred_cal_gen)    
                    
                    # calculating accuracies for optimized model
                    self.model_best.fit(self.x[:,self.best_features],self.y)
                    self.y_pred_cal_op = np.squeeze(self.model_best.predict(self.x[:,self.best_features])).squeeze()
                    self.r2_cal_op = r2_score(self.y,self.y_pred_cal_op)
                    self.r2_adj_cal_op = adj_r2_score(self.y,x=self.x[:,self.best_features], r2=self.r2_cal_op)
                    self.rmse_cal_op = mean_squared_error(self.y, self.y_pred_cal_op, squared = False)  
                    self.rmspe_cal_op = rmspe_score(self.y, self.y_pred_cal_op)    
                    
                    # generalized model
                    fig,axes = plt.subplots(1,3,figsize=(15,5))
                    sns.regplot(x=self.y, y=self.y_pred_cal_gen, color="k",ax=axes[0])
                    axes[0].plot([np.min([self.y, self.y_pred_cal_gen]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cal_gen]) + 0.1*np.std(self.y)],
                        [np.min([self.y, self.y_pred_cal_gen]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cal_gen]) + 0.1*np.std(self.y)],'r--')
                    axes[0].text(0.05,0.80,r"RMSE={:.3f}".format(self.rmse_cal_gen) + "\n"+ 
                                     r"RMSPE={:.3f}".format(self.rmspe_cal_gen) + "\n"+ 
                                      r"$R^2$={:.3f}".format(self.r2_cal_gen)+ "\n"+
                                      r"$Adj. R^2$={:.3f}".format(self.r2_adj_cal_gen),
                                      transform=axes[0].transAxes,
                                      color='k',
                                      bbox=dict(facecolor=(1,1,1,0.5),edgecolor=(0,0,0,1)))
                    axes[0].set_xlabel("Measured")
                    axes[0].set_ylabel("Predicted")
                    axes[0].set_title("General model - Cal.")
                    axes[0].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                                                          
                    
                    
                    # optimized model
                    sns.regplot(x=self.y, y=self.y_pred_cal_op, color="k",ax=axes[1])
                    axes[1].text(0.05,0.80,r"RMSE={:.3f}".format(self.rmse_cal_op) + "\n"+ 
                                     r"RMSPE={:.3f}".format(self.rmspe_cal_op) + "\n"+ 
                                      r"$R^2$={:.3f}".format(self.r2_cal_op)+ "\n"+
                                      r"$Adj. R^2$={:.3f}".format(self.r2_adj_cal_op),
                                      transform=axes[1].transAxes,
                                      color='k',
                                      bbox=dict(facecolor=(1,1,1,0.5),edgecolor=(0,0,0,1)))
                    axes[1].plot([np.min([self.y, self.y_pred_cal_op]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cal_op]) + 0.1*np.std(self.y)],
                        [np.min([self.y, self.y_pred_cal_op]) - 0.1*np.std(self.y), np.max([self.y, self.y_pred_cal_op]) + 0.1*np.std(self.y)],'r--')
                    axes[1].set_xlabel("Measured")
                    axes[1].set_ylabel("Predicted")
                    axes[1].set_title("Optimized model - Cal.")
                    axes[1].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                                       
                    plt.show(block=False)
                    
                    sns.residplot(self.y_pred_cal_op, self.y - self.y_pred_cal_op, lowess=False, color="r",ax=axes[2])
                    axes[2].set_xlabel("Predicted")
                    axes[2].set_ylabel("Residuals")
                    axes[2].set_title("Optimized model - Cal. Residuals") 
                    axes[2].ticklabel_format(axis="both", style = "sci", scilimits = (0,2))                                                       

                    results = {"r2_cal_gen":[self.r2_cal_gen],'rmse_cal_gen':[self.rmse_cal_gen],'rmspe_cal_gen':[self.rmspe_cal_gen],"r2_adj_cal_gen":[self.r2_adj_cal_gen],
                                   "r2_cal_op":[self.r2_cal_op],'rmse_cal_op':[self.rmse_cal_op],'rmspe_cal_op':[self.rmspe_cal_op], "r2_adj_cal_op":[self.r2_adj_cal_op]}                          
            
            # if the task is classification
            else:
                if hasattr(self, 'x_train') and hasattr(self, 'x_test') and hasattr(self, 'y_train') and hasattr(self, 'y_test'):         
                    if len(self.class_sep) == 2:
                        fig,axes = plt.subplots(2,3,figsize=(15,10))
                    else:
                        fig,axes = plt.subplots(2,2,figsize=(10 + len(self.class_sep),10 + len(self.class_sep)))
                    # general model
                    self.model.fit(self.x_train,self.y_train)
                    y_train_pred  = self.model.predict(self.x_train)
                    self.acc_train_gen = self.model.score(self.x_train,self.y_train)
                    cm = confusion_matrix(self.y_train,y_train_pred.squeeze())
                    cm = np.stack(list(map(lambda x: x/x.sum(),cm)))
                    conf_matrix=pd.DataFrame(cm,columns = self.class_sep,index= self.class_sep)
                    sns.heatmap(conf_matrix,cmap="coolwarm", annot=True, ax=axes[0,0],vmin=0,vmax=1)
                    axes[0,0].set_title('General model - Cal. confusion matrix')
                    axes[0,0].set_ylabel('True label')
                    axes[0,0].set_xlabel('Predicted label')

                    self.model.fit(self.x_train,self.y_train)
                    y_test_pred  = self.model.predict(self.x_test)
                    self.acc_test_gen = self.model.score(self.x_test,self.y_test)
                    cm = confusion_matrix(self.y_test,y_test_pred.squeeze())
                    cm = np.stack(list(map(lambda x: x/x.sum(),cm)))
                    conf_matrix=pd.DataFrame(cm,columns = self.class_sep,index= self.class_sep)
                    sns.heatmap(conf_matrix,cmap="coolwarm", annot=True,ax = axes[1,0],vmin=0,vmax=1)
                    axes[1,0].set_title('General model - Val. confusion matrix')  
                    axes[1,0].set_ylabel('True label')
                    axes[1,0].set_xlabel('Predicted label')
        
                    # optimized model
                    self.model_best.fit(self.x_train[:,self.best_features],self.y_train)
                    y_train_pred  = self.model_best.predict(self.x_train[:,self.best_features])
                    self.acc_train_op = self.model_best.score(self.x_train[:,self.best_features],self.y_train)
                    cm = confusion_matrix(self.y_train,y_train_pred.squeeze())
                    cm = np.stack(list(map(lambda x: x/x.sum(),cm)))
                    conf_matrix=pd.DataFrame(cm,columns = self.class_sep,index= self.class_sep)
                    sns.heatmap(conf_matrix,cmap="coolwarm", annot=True, ax=axes[0,1],vmin=0,vmax=1)
                    axes[0,1].set_title('Optimized model - Cal. confusion matrix')     
                    axes[0,1].set_ylabel('True label')
                    axes[0,1].set_xlabel('Predicted label')                
                    
                    self.model_best.fit(self.x_train[:,self.best_features],self.y_train)
                    y_test_pred  = self.model_best.predict(self.x_test[:,self.best_features])
                    self.acc_test_op = self.model_best.score(self.x_test[:,self.best_features],self.y_test)
                    cm = confusion_matrix(self.y_test,y_test_pred.squeeze())
                    cm = np.stack(list(map(lambda x: x/x.sum(),cm)))
                    conf_matrix=pd.DataFrame(cm,columns = self.class_sep,index= self.class_sep)
                    sns.heatmap(conf_matrix,cmap="coolwarm", annot=True,ax = axes[1,1],vmin=0,vmax=1)
                    axes[1,1].set_title('Optimized model - Val. confusion matrix')      
                    axes[1,1].set_ylabel('True label')
                    axes[1,1].set_xlabel('Predicted label')
        
                    if len(self.class_sep) == 2:
                        #optimized model Roc curves
                        self.model_best.fit(self.x_train[:,self.best_features],self.y_train)
                        y_train_pred_prob = self.model_best.predict_proba(self.x_train[:,self.best_features])
                        fpr, tpr, thr = roc_curve(self.y_train, y_train_pred_prob[:, 1].squeeze())   
                        self.auc_train_op = auc(fpr, tpr)
                        axes[0,2].plot([0,1],[0,1],linestyle = '--',lw = 2,color = 'black')
                        axes[0,2].plot(fpr, tpr, color='blue',
                                 label=r'$ROC_{train}$ (AUC = %0.2f)' % (auc(fpr, tpr)),lw=2, alpha=1)
                        axes[0,2].set_xlabel('False Positive Rate')
                        axes[0,2].set_ylabel('True Positive Rate')
                        axes[0,2].set_title('ROC Curve')
                        axes[0,2].legend(loc="lower right")
                        
                        self.model_best.fit(self.x_train[:,self.best_features],self.y_train)
                        y_test_pred_prob = self.model_best.predict_proba(self.x_test[:,self.best_features])
                        fpr, tpr, thr = roc_curve(self.y_test, y_test_pred_prob[:, 1].squeeze())      
                        self.auc_test_op = auc(fpr, tpr)
                        axes[1,2].plot([0,1],[0,1],linestyle = '--',lw = 2,color = 'black')
                        axes[1,2].plot(fpr, tpr, color='blue',
                                 label=r'$ROC_{test}$ (AUC = %0.2f)' % (auc(fpr, tpr)),lw=2, alpha=1)
                        axes[1,2].set_xlabel('False Positive Rate')
                        axes[1,2].set_ylabel('True Positive Rate')
                        axes[1,2].set_title('ROC Curve')
                        axes[1,2].legend(loc="lower right")
                    
                        results = {"acc_train_gen":[self.acc_train_gen],'acc_test_gen':[self.acc_test_gen],
                           'acc_train_op':[self.acc_train_op],'acc_test_op':[self.acc_test_op],
                           "auc_train_op":[self.auc_train_op],'auc_test_op':[self.auc_test_op]}
                    
                    results = {"acc_train_gen":[self.acc_train_gen],'acc_test_gen':[self.acc_test_gen],
                       'acc_train_op':[self.acc_train_op],'acc_test_op':[self.acc_test_op]}
                
                # if train and test not defined
                else:
                    warnings.warn("Neither cv_model nor test partition (x_test/y_test) are defined, predicting on x/y.")
                    if len(self.class_sep) == 2:
                        fig,axes = plt.subplots(1,3,figsize=(15,5))
                    else:
                        fig,axes = plt.subplots(1,2,figsize=(10 + len(self.class_sep),5 + len(self.class_sep)))                        
                    # general model
                    self.model.fit(self.x,self.y)
                    y_train_pred  = self.model.predict(self.x)
                    self.acc_train_gen = self.model.score(self.x,self.y)
                    cm = confusion_matrix(self.y,y_train_pred.squeeze())
                    cm = np.stack(list(map(lambda x: x/x.sum(),cm)))
                    conf_matrix=pd.DataFrame(cm,columns = self.class_sep,index= self.class_sep)
                    sns.heatmap(conf_matrix,cmap="coolwarm", annot=True, ax=axes[0],vmin=0,vmax=1)
                    axes[0].set_title('General model - Cal. confusion matrix')
                    axes[0].set_ylabel('True label')
                    axes[0].set_xlabel('Predicted label')
        
                    # optimized model
                    self.model_best.fit(self.x[:,self.best_features],self.y)
                    y_train_pred  = self.model_best.predict(self.x[:,self.best_features])
                    self.acc_train_op = self.model_best.score(self.x[:,self.best_features],self.y)
                    cm = confusion_matrix(self.y,y_train_pred.squeeze())
                    cm = np.stack(list(map(lambda x: x/x.sum(),cm)))
                    conf_matrix=pd.DataFrame(cm,columns = self.class_sep,index= self.class_sep)
                    sns.heatmap(conf_matrix,cmap="coolwarm", annot=True, ax=axes[1],vmin=0,vmax=1)
                    axes[1].set_title('Optimized model - Cal. confusion matrix')     
                    axes[1].set_ylabel('True label')
                    axes[1].set_xlabel('Predicted label')                
        
        
                    if len(self.class_sep) == 2:
                        #optimized model Roc curves
                        self.model_best.fit(self.x[:,self.best_features],self.y)
                        y_train_pred_prob = self.model_best.predict_proba(self.x[:,self.best_features])
                        fpr, tpr, thr = roc_curve(self.y, y_train_pred_prob[:, 1].squeeze())   
                        self.auc_train_op = auc(fpr, tpr)
                        axes[2].plot([0,1],[0,1],linestyle = '--',lw = 2,color = 'black')
                        axes[2].plot(fpr, tpr, color='blue',
                                 label=r'$ROC_{train}$ (AUC = %0.2f)' % (auc(fpr, tpr)),lw=2, alpha=1)
                        axes[2].set_xlabel('False Positive Rate')
                        axes[2].set_ylabel('True Positive Rate')
                        axes[2].set_title('ROC Curve')
                        axes[2].legend(loc="lower right")
    
                    
                        results = {"acc_train_gen":[self.acc_train_gen],
                           'acc_train_op':[self.acc_train_op],
                           "auc_train_op":[self.auc_train_op]}
                    
                    results = {"acc_train_gen":[self.acc_train_gen],
                       'acc_train_op':[self.acc_train_op]}
                    
        plt.show(block=False) 
        return results, fig
      
    def _decorrellate_thr(self,x_df, threshold):        
        dataset = x_df.copy()
        col_corr = set() # Set of all the names of deleted columns
        corr_matrix = dataset.corr()
        for i in range(len(corr_matrix.columns)):
            for j in range(i):
                if (np.abs(corr_matrix.iloc[i, j]) >= threshold) and (corr_matrix.columns[j] not in col_corr):
                    colname = corr_matrix.columns[i] # getting the name of column
                    col_corr.add(colname)
                    if colname in dataset.columns:
                        del dataset[colname] # deleting the column from the dataset
        new_dataset = dataset.copy()
        return new_dataset        

    def _decorrellate_n(self,x_df, n):        
        # optimizing the decorrellate_thresh to find the threshold that gives out number of features matching with that
        x0 = [0.9]
        func = lambda x: np.abs(len(self._decorrellate_thr(x_df,x).columns) - n) #minimizng this function 
        thresh = minimize(func, x0, method = "Nelder-Mead").x[0]        
        new_dataset = self._decorrellate_thr(x_df,thresh)
        self.decor_thresh = thresh        
        return new_dataset              
    
    def _scorer_maker(self,**kwargs):
        scorer_func = partial(self.scoring,**kwargs)
        update_wrapper(scorer_func, self.scoring)
        return scorer_func
    
    def _feature_importance(self,approach):
        
        valid_approaches = ["variance",
                            "f_test",
                            "mutual_info",
                            "chi2",
                            "pearson"]
        if approach not in valid_approaches:
            raise ValueError(f"Valid approaches are {valid_approaches}.")
    
        if approach == "variance":
            selector = VarianceThreshold()
            selector.fit(self.x)
            var = selector.variances_
            indices = np.argsort(var)[::-1]
            
        elif approach == "f_test":
            if is_regressor(self.model):
                f,_ = f_regression(self.x,self.y)
                indices = np.argsort(f)[::-1]
            else:
                f,_ = f_classif(self.x,self.y)
                indices = np.argsort(f)[::-1]
                            
        elif approach == "mutual_info":
            if is_regressor(self.model):
                mi = mutual_info_regression(self.x,self.y)
                indices = np.argsort(mi)[::-1]
            else:
                mi = mutual_info_classif(self.x,self.y)
                indices = np.argsort(mi)[::-1]            
            
        elif approach == "chi2":            
            if is_regressor(self.model):
                raise ValueError("chi2 is only available for classification.")
            else:
                chi_sqaured,_ = chi2(self.x,self.y)
                indices = np.argsort(chi_sqaured)[::-1]            
                
        elif approach == "pearson":
            if is_regressor(self.model):            
                r = np.abs(np.array([stats.pearsonr(self.x[:,i],self.y)[0] for i in range(self.x.shape[1])]))
                indices = np.argsort(r)[::-1]        
            else:
                raise ValueError("pearson is only available for regression.")
                            
        return indices
    
    def _feature_ranking(self, tops,columns, n_repeats, random_state):
        # here we are multiplying the initial permutation feature importance by the weight acquired
        # from the top 25% percentile agents and their occurence ratio
        top_agents, top_perc = tops
        
        # getting w        
        top_agents_arr = np.concatenate(top_agents)
        indices, counts = np.unique(top_agents_arr,return_counts=True) #indices are all features in the pool, and counts is the number of appearance totally
        w = counts/len(top_agents)
        
        # getting alpha, which is based on percentage score (rmspe or f1)
        # first creating a mask to see where features belong
        indices_mask = np.zeros((len(indices),len(top_agents)),dtype=bool)        
        for c1, index in enumerate(indices):    
            for c2, agent in enumerate(top_agents):
                indices_mask[c1,c2] = index in agent 
                
        # getting mean perc based on the mask for each index (feature)        
        alpha = np.zeros(w.shape)
        for c, index_mask in enumerate(indices_mask):
            alpha[c] = np.mean(np.array(top_perc)[index_mask])                    
        
        model = self.model_best.fit(self.x,self.y)
        rankings = permutation_importance(model, self.x, self.y,n_repeats=n_repeats,random_state=random_state).importances_mean        
        new_rankings = copy(rankings)
                
        for c,index in enumerate(indices):    
            if is_regressor(model): #using rmspe
                new_rankings[index] = new_rankings[index]*(1 + (1-alpha[c])*w[c])
            else: #using f1
                new_rankings[index] = new_rankings[index]*(1 + alpha[c]*w[c])
        
        new_rankings = np.concatenate([np.expand_dims(columns,0), 
                                       np.expand_dims(new_rankings,0)],axis=0)
        
        return new_rankings 
         
    def _check_scoring(self):
        raise DeprecationWarning("The function is deprecated and will be removed in future versions. _scorer_maker replaces this function.")
        
        # setting up scoring
        score_valid = {
        # classification accuracies
        "accuracy": accuracy_score,
        "balanced_accuracy":balanced_accuracy_score,
        "average_precision":average_precision_score,
        "f1":f1_score,
        "neg_brier_score":brier_score_loss,
        "neg_log_loss":log_loss,
        "precision":precision_score,
        "recall":recall_score,
        "jaccard":jaccard_score,
        "kappa":kappa_score,
        # regression accuracies 
        "explained_variance":explained_variance_score,
        "max_error":max_error,
        "mean_absolute_error":mean_absolute_error,
        "mean_squared_error":mean_squared_error,
        "root_mean_squared_error":rmse_score,
        "mean_squared_log_error":mean_squared_log_error,
        "median_absolute_error":median_absolute_error,
        "r2":r2_score,
        "adj_r2":adj_r2_score,
        "mean_poisson_deviance":mean_poisson_deviance,
        "mean_gamma_deviance":mean_gamma_deviance,
        "mean_tweedie_deviance":mean_tweedie_deviance}
        
        
        # the function takes in a string and returns a scorer
        if self.scoring in score_valid.keys():
            scorer = score_valid[self.scoring]
        else:
            raise ValueError(f"{self.scoring} is not a valid scorer. Valid scorers are:\n {score_valid.keys()}")
                
        return scorer    
         
        
        
        
        
        
        
        
        
        
        
        
        
        
        




# -*- coding: utf-8 -*-
"""
Created on Mon Dec 14 13:19:12 2020

@author: Amirh
"""
from hyperopt import hp
from hyperopt import fmin, tpe, space_eval

#%%

class ParamTuner():
    def __init__(self, opt_model,
                 x,y,
                 decor=None, scale=False, 
                 cv=None, test_size=None):
        """
        The parameter tuning class using Hyperopt module.

        Parameters
        ----------
        opt_model : Class
            The instantiated class of optimization model.
        x : ndarray or sparse matrix of shape (n_samples, n_features)
            Input data.
        y : ndarray of shape (n_samples,)
            Target values.
        decor : tuple or float, optional
            Decorrellation parameter. If a number in range [0,1] is chosen
            then the input features that have a correllation factor above the 
            give value, are removed. If tuple of (string, float [0,1]) is chosen
            then the input data features are ranked based on their feature
            importance method, indicated by the string, then decorrelated and the
            ones with higher importance are retained. The string could be one
            from the following:                
                "variance"
                "f_test"
                "mutual_info"
                "chi2"
                "pearson"                
            The default is None.
        scale : bool, optional
            Wether to scale the input data or not (centering and scaling).
            The default is False.
        test_size : float, optional
            Represents the portion of input data to be included for test split. 
            Needs to be between 0 and 1. The default is None.

        """
        
        self.x = x
        self.y = y
        self.opt_model = opt_model # TODO make sure all the otpimizers have best_sol_acc, and make sure you could also pass NSGA2 to this
        self.cv = cv
        self.decor = decor
        self.scale = scale
        self.test_size = test_size

    def tune(self, param_dict, max_evals=1000, index = 0, obj = 0):
        """
        Function taking in the dictionary of parameters and performing 
        hyperopt tuning. 

        Parameters
        ----------
        param_dict : Dictionary
            Dictionary of parameter for the optimization algorithms.
        max_evals : int, optional
            Maximum number of evaluation. The default is 1000.
        index : int, optional
            index of the accurcaies for the NSGA2 results. If NSGA2 is being
            optimized and nothing is passed, it will choose the first index.
        obj : int, optional
            index of the objective for the NSGA2 results. If nothing is passed, first one is chosen.
        Returns
        -------
        best : Dataframe
            Dataframe of best parameters.
        """
        self.index = index
        self.obj = obj
        best = fmin(self._objective, param_dict, algo=tpe.suggest, max_evals=max_evals)
        return best 
    
    def _objective(self,args):          
        self.opt_model.__dict__.update((k, v) for k, v in args.items())    
        
        # # optimizing 
        self.opt_model.fit(self.x,self.y,decor = self.decor, scale = self.scale, test_size = self.test_size)
        if self.opt_model._name_ == "NSGA2":
            best_fit = self.opt_model.best_sols_acc[self.obj][self.index]
        else:
            best_fit = self.opt_model.best_sol_acc
        
        if self.opt_model._name_ == "NSGA2":
            if self.opt_model.weight[self.obj] == 1:
                return - best_fit
            elif self.opt_model.weight[self.obj] == -1:
                return best_fit
        else:            
            if self.opt_model.weight == 1:
                return - best_fit
            elif self.opt_model.weight == -1:
                return best_fit
    


# -*- coding: utf-8 -*-
"""
Created on Mon Dec 14 13:19:12 2020

@author: Amirh
"""
from algorithms._aco_v2 import ACO
import numpy as np
import pandas as pd
from sklearn.cross_decomposition import PLSRegression
from sklearn.model_selection import KFold
from sklearn.model_selection import RandomizedSearchCV
from scipy.stats import loguniform
from functions._metrics import adj_r2_score
from functions.decor import decor
from hyperopt import hp
from hyperopt import fmin, tpe, space_eval
from sklearn.linear_model import LinearRegression
from scipy.stats import randint
from sklearn.model_selection import LeaveOneOut
#%%
def objective(args):  
           
    nAnt = args["nAnt"]        
    Q = args["Q"]        
    tau0 = args["tau0"]          
    beta = args["beta"]     
    rho = args["rho"]     
    
    data = pd.read_csv(r"C:\Users\Amirh\Downloads\sfm_metrics_28mflight.csv").to_numpy()
    x = data[:,2:]
    y = data[:,1]
    model = PLSRegression()
    #cv = KFold()
    cv = None
    
    # # optimizing 
    aco = ACO(model=model, n_f=5, weight=1, scoring='r2', cv=cv , n_iter=20, nAnt = nAnt, Q=Q, tau0=tau0, 
                      alpha=1, beta=beta, rho=rho,random_state=42)

    aco.load(x,y, decor = 0.9, scale = True)
    aco.optimize()
    best_fit = aco.best_sol_acc
    
    return -best_fit
#%%


space = {
        'nAnt': hp.randint('nAnt', 50,200),
        "Q":  hp.loguniform("Q",np.log(1e-2),np.log(0.99)),        
        "tau0":  hp.loguniform("tau0",np.log(1e-2),np.log(0.99)),        
        "beta": hp.randint('beta', 2,5),
        "rho": hp.loguniform('rho', np.log(1e-2),np.log(0.99))
        }

best = fmin(objective, space, algo=tpe.suggest, max_evals=100)
#%%


data = pd.read_csv(r"C:\Users\Amirh\Downloads\sfm_metrics_28mflight.csv").to_numpy()
x = data[:,2:]
y = data[:,1]
model = PLSRegression()
cv = None
#cv = LeaveOneOut()
#cv = KFold(10)
 
# # optimizing 
aco = ACO(model=model, n_f=5, weight=1, scoring='r2', cv=cv , n_iter=1000, nAnt = 124, Q=0.02401580548214845, tau0=0.21201587723743645, 
                  alpha=1, beta=4, rho=0.5585333355449467,random_state=42)
aco.load(x,y,decor = 0.9, scale = True)
aco.optimize()

param_dict = {"n_components": randint(1,5)}
aco.tune_model_params(param_dict)
aco.display_results()

"""
0.7868041189260054

beta 4
nAnt 124
Q 0.02401580548214845
rho 0.5585333355449467
tau0 0.21201587723743645
"""


# -*- coding: utf-8 -*-
"""
Created on Wed Mar 31 14:57:23 2021

@author: Amirh
"""

from sklearn.datasets import make_classification, make_regression
from sklearn.metrics import r2_score
from jostar.algorithms import ACO, GA, SA, PSO, PlusLMinusR, DE, NSGA2, SBS, SFS
from sklearn.svm import SVR, SVC
from sklearn.base import is_classifier, is_regressor
from sklearn.model_selection import KFold
import warnings
import matplotlib.pyplot as plt
import pandas as pd
from tqdm import tqdm
import matplotlib



def eval_opt_model_output_regression(opt_model, n_f):
    rank_models = ["GA", "SA", "PSO", "ACO", "DE"]
    seq_models = ["PlusLMinusR", "SFS", "SBS"]
    if opt_model._name_ in rank_models:
        assert len(opt_model.best_fits) == 1
        assert len(opt_model.best_sol) == n_f
        assert is_regressor(opt_model.model_best)
        assert len(opt_model.rankings) == 2
        assert opt_model.display_results() is not None
    elif opt_model._name_ in seq_models:
        if opt_model._name_ != "SBS":
            assert len(opt_model.best_fits) == n_f
            assert len(opt_model.best_sol) == n_f
            assert is_regressor(opt_model.model_best)
            assert opt_model.display_results() is not None
        else:
            assert len(opt_model.best_sol) == n_f
            assert is_regressor(opt_model.model_best)
            assert opt_model.display_results() is not None
    else:
        result_df = opt_model.res_df
        assert isinstance(result_df, pd.core.frame.DataFrame)
        assert opt_model.display_results(0) is not None
    plt.close('all')
    pbar.update(1)


def eval_opt_model_output_classification(opt_model, n_f):
    rank_models = ["GA", "SA", "PSO", "ACO", "DE"]
    seq_models = ["PlusLMinusR", "SFS", "SBS"]
    if opt_model._name_ in rank_models:
        assert len(opt_model.best_fits) == 1
        assert len(opt_model.best_sol) == n_f
        assert is_classifier(opt_model.model_best)
        rank_models = ["GA", "SA", "PSO", "ACO", "DE"]
        if opt_model._name_ in rank_models:
            assert len(opt_model.rankings) == 2
        assert opt_model.display_results() is not None
    elif opt_model._name_ in seq_models:
        if opt_model._name_ != "SBS":
            assert len(opt_model.best_fits) == n_f
            assert len(opt_model.best_sol) == n_f
            assert is_classifier(opt_model.model_best)
            assert opt_model.display_results() is not None
        else:
            assert len(opt_model.best_sol) == n_f
            assert is_classifier(opt_model.model_best)
            assert opt_model.display_results() is not None
    else:
        result_df = opt_model.res_df
        assert isinstance(result_df, pd.core.frame.DataFrame)
        assert opt_model.display_results(0) is not None
    plt.close('all')
    pbar.update(1)


def test_all_regression():
    global pbar
    cv = KFold(5)
    n_f = 5
    x, y = make_regression(100, 10)
    model = SVR()

    # regression
    # with CV
    pbar = tqdm(total=18)
    ga_opt_model = GA(model, n_f, +1, r2_score, n_gen=1,
                      n_pop=20, cv=cv, verbose=False)
    sa_opt_model = SA(model, n_f, +1, r2_score, n_iter=1,
                      n_sub_iter=20, cv=cv, verbose=False)
    de_opt_model = DE(model, n_f, +1, r2_score, n_iter=1,
                      n_pop=20, cv=cv, verbose=False)
    aco_opt_model = ACO(model, n_f, +1, r2_score, n_iter=1,
                        n_ant=20, cv=cv, verbose=False)
    pso_opt_model = PSO(model, n_f, +1, r2_score, n_iter=1,
                        n_pop=20, cv=cv, verbose=False)
    lrs_opt_model = PlusLMinusR(model, n_f, +1, r2_score, cv=cv, verbose=False)
    nsga_opt_model = NSGA2(model, n_f, (+1, -1), r2_score,
                           n_gen=1, n_pop=20, cv=cv, verbose=False)
    sbs_opt_model = SBS(model, n_f, +1, r2_score, cv=cv, verbose=False)
    sfs_opt_model = SFS(model, n_f, +1, r2_score, cv=cv, verbose=False)

    ga_opt_model.fit(x, y, decor=0.95, scale=True)
    sa_opt_model.fit(x, y, decor=0.95, scale=True)
    de_opt_model.fit(x, y, decor=0.95, scale=True)
    aco_opt_model.fit(x, y, decor=0.95, scale=True)
    pso_opt_model.fit(x, y, decor=0.95, scale=True)
    lrs_opt_model.fit(x, y, decor=0.95, scale=True)
    nsga_opt_model.fit(x, y, decor=0.95, scale=True)
    sbs_opt_model.fit(x, y, decor=0.95, scale=True)
    sfs_opt_model.fit(x, y, decor=0.95, scale=True)

    eval_opt_model_output_regression(ga_opt_model, n_f)
    eval_opt_model_output_regression(sa_opt_model, n_f)
    eval_opt_model_output_regression(de_opt_model, n_f)
    eval_opt_model_output_regression(aco_opt_model, n_f)
    eval_opt_model_output_regression(pso_opt_model, n_f)
    eval_opt_model_output_regression(lrs_opt_model, n_f)
    eval_opt_model_output_regression(nsga_opt_model, n_f)
    eval_opt_model_output_regression(sbs_opt_model, n_f)
    eval_opt_model_output_regression(sfs_opt_model, n_f)

    # with test size
    ga_opt_model = GA(model, n_f, +1, r2_score, n_gen=1,
                      n_pop=20, cv=None, verbose=False)
    sa_opt_model = SA(model, n_f, +1, r2_score, n_iter=1,
                      n_sub_iter=20, cv=None, verbose=False)
    de_opt_model = DE(model, n_f, +1, r2_score, n_iter=1,
                      n_pop=20, cv=None, verbose=False)
    aco_opt_model = ACO(model, n_f, +1, r2_score, n_iter=1,
                        n_ant=20, cv=None, verbose=False)
    pso_opt_model = PSO(model, n_f, +1, r2_score, n_iter=1,
                        n_pop=20, cv=None, verbose=False)
    lrs_opt_model = PlusLMinusR(
        model, n_f, +1, r2_score, cv=None, verbose=False)
    nsga_opt_model = NSGA2(model, n_f, (+1, -1), r2_score,
                           n_gen=1, n_pop=20, cv=None, verbose=False)
    sbs_opt_model = SBS(model, n_f, +1, r2_score, cv=None, verbose=False)
    sfs_opt_model = SFS(model, n_f, +1, r2_score, cv=None, verbose=False)

    ga_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    sa_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    de_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    aco_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    pso_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    lrs_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    nsga_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    sbs_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    sfs_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)

    eval_opt_model_output_regression(ga_opt_model, n_f)
    eval_opt_model_output_regression(sa_opt_model, n_f)
    eval_opt_model_output_regression(de_opt_model, n_f)
    eval_opt_model_output_regression(aco_opt_model, n_f)
    eval_opt_model_output_regression(pso_opt_model, n_f)
    eval_opt_model_output_regression(lrs_opt_model, n_f)
    eval_opt_model_output_regression(nsga_opt_model, n_f)
    eval_opt_model_output_regression(sbs_opt_model, n_f)
    eval_opt_model_output_regression(sfs_opt_model, n_f)


def test_all_classification():
    global pbar
    cv = KFold(5)
    n_f = 5
    x, y = make_classification(100, 10)
    model = SVC(probability=True)

    # regression
    # with CV
    pbar = tqdm(total=18)
    ga_opt_model = GA(model, n_f, +1, r2_score, n_gen=1,
                      n_pop=20, cv=cv, verbose=False)
    sa_opt_model = SA(model, n_f, +1, r2_score, n_iter=1,
                      n_sub_iter=20, cv=cv, verbose=False)
    de_opt_model = DE(model, n_f, +1, r2_score, n_iter=1,
                      n_pop=20, cv=cv, verbose=False)
    aco_opt_model = ACO(model, n_f, +1, r2_score, n_iter=1,
                        n_ant=20, cv=cv, verbose=False)
    pso_opt_model = PSO(model, n_f, +1, r2_score, n_iter=1,
                        n_pop=20, cv=cv, verbose=False)
    lrs_opt_model = PlusLMinusR(model, n_f, +1, r2_score, cv=cv, verbose=False)
    nsga_opt_model = NSGA2(model, n_f, (+1, -1), r2_score,
                           n_gen=1, n_pop=20, cv=cv, verbose=False)
    sbs_opt_model = SBS(model, n_f, +1, r2_score, cv=cv, verbose=False)
    sfs_opt_model = SFS(model, n_f, +1, r2_score, cv=cv, verbose=False)

    ga_opt_model.fit(x, y, decor=0.95, scale=True)
    sa_opt_model.fit(x, y, decor=0.95, scale=True)
    de_opt_model.fit(x, y, decor=0.95, scale=True)
    aco_opt_model.fit(x, y, decor=0.95, scale=True)
    pso_opt_model.fit(x, y, decor=0.95, scale=True)
    lrs_opt_model.fit(x, y, decor=0.95, scale=True)
    nsga_opt_model.fit(x, y, decor=0.95, scale=True)
    sbs_opt_model.fit(x, y, decor=0.95, scale=True)
    sfs_opt_model.fit(x, y, decor=0.95, scale=True)

    eval_opt_model_output_classification(ga_opt_model, n_f)
    eval_opt_model_output_classification(sa_opt_model, n_f)
    eval_opt_model_output_classification(de_opt_model, n_f)
    eval_opt_model_output_classification(aco_opt_model, n_f)
    eval_opt_model_output_classification(pso_opt_model, n_f)
    eval_opt_model_output_classification(lrs_opt_model, n_f)
    eval_opt_model_output_classification(nsga_opt_model, n_f)
    eval_opt_model_output_classification(sbs_opt_model, n_f)
    eval_opt_model_output_classification(sfs_opt_model, n_f)

    # with test size
    ga_opt_model = GA(model, n_f, +1, r2_score, n_gen=1,
                      n_pop=20, cv=None, verbose=False)
    sa_opt_model = SA(model, n_f, +1, r2_score, n_iter=1,
                      n_sub_iter=20, cv=None, verbose=False)
    de_opt_model = DE(model, n_f, +1, r2_score, n_iter=1,
                      n_pop=20, cv=None, verbose=False)
    aco_opt_model = ACO(model, n_f, +1, r2_score, n_iter=1,
                        n_ant=20, cv=None, verbose=False)
    pso_opt_model = PSO(model, n_f, +1, r2_score, n_iter=1,
                        n_pop=20, cv=None, verbose=False)
    lrs_opt_model = PlusLMinusR(
        model, n_f, +1, r2_score, cv=None, verbose=False)
    nsga_opt_model = NSGA2(model, n_f, (+1, -1), r2_score,
                           n_gen=1, n_pop=20, cv=None, verbose=False)
    sbs_opt_model = SBS(model, n_f, +1, r2_score, cv=None, verbose=False)
    sfs_opt_model = SFS(model, n_f, +1, r2_score, cv=None, verbose=False)

    ga_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    sa_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    de_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    aco_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    pso_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    lrs_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    nsga_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    sbs_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)
    sfs_opt_model.fit(x, y, decor=0.95, scale=True, test_size=0.3)

    eval_opt_model_output_classification(ga_opt_model, n_f)
    eval_opt_model_output_classification(sa_opt_model, n_f)
    eval_opt_model_output_classification(de_opt_model, n_f)
    eval_opt_model_output_classification(aco_opt_model, n_f)
    eval_opt_model_output_classification(pso_opt_model, n_f)
    eval_opt_model_output_classification(lrs_opt_model, n_f)
    eval_opt_model_output_classification(nsga_opt_model, n_f)
    eval_opt_model_output_classification(sbs_opt_model, n_f)
    eval_opt_model_output_classification(sfs_opt_model, n_f)


if __name__ == '__main__':
    test_all_regression()
    test_all_classification()
    print(" ")
    print("All tests passed!")



# -*- coding: utf-8 -*-
"""
Created on Thu Sep 24 14:58:43 2020

@author: Amirh
"""
from ..utils._base_feature_selector import BaseFeatureSelector
from ..functions import ff
import numpy as np
import random
from sklearn.base import clone
import time

class _DifferentialEvolution():
    def __init__(self, estimator,cv, scorer, nf, fitness, weight,
                 x,y,x_test= None, y_test = None,
                 MaxIt = 1000, nPop = 10, 
                 low_bound = 0.5, upper_bound = 1, cross_proba = 0.3,
                 random_state=None,verbose = True):
        
        self.nf = nf
        self.estimator = estimator
        self.cv = cv
        self.scorer = scorer
        self.x = x
        self.y = y
        self.x_test = x_test
        self.y_test = y_test
        self.fitness = fitness
        self.weight = weight    
        self.MaxIt = MaxIt
        self.nPop = nPop    
        
        self.VarSize = self.x.shape[1]
        self.VarMin = 0
        self.VarMax = 1
        
        # DE parameters
        self.beta_min = low_bound #lower bound scaling factor
        self.beta_max = upper_bound #upper bound scaling factor
        self.pCR = cross_proba      #crossover probability
        
        self.verbose = verbose
        self.random_state = random_state
        self.scorer = scorer
        
        random.seed(self.random_state)        
        np.random.seed(self.random_state)        

    def RK(self,u):
        # this function creates random keys based on continuous values
        return np.argsort(u)[:self.nf]
        #return np.argsort(u)
    
    def Mutation(self,a,b,c):
        beta = np.random.uniform(self.beta_min,self.beta_max,self.VarSize)
        mutated = self.pop[a].position + np.power(beta,self.pop[b].position - self.pop[c].position)
        return mutated
    
    def CrossOver(self,ind1,ind2):
        ind3 = np.zeros(len(ind1))
        j0 = random.randint(0,len(ind1)-1)
        
        for j in range(len(ind1)):
            if j == j0 and np.random.uniform(0, 1) <= self.pCR:
                ind3[j] = ind2[j]
            else:
                ind3[j] = ind1[j]
                
        return ind3
    
    def FeatureRanker(self,all_agents):
        agents = [self.RK(agent.position) for agent in all_agents]
        fits = [agent.fit for agent in all_agents]
        percs = [agent.perc for agent in all_agents]
        
        if self.weight == 1:
            fits_first = np.percentile(fits,75)
            fits_second = np.percentile(fits,100)        
        elif self.weight == -1:
            fits_first = np.percentile(fits,0)
            fits_second = np.percentile(fits,25)        
            
        indices = np.where(np.logical_and((fits >= fits_first), (fits <= fits_second)))[0]
        
        top_agents = np.array(agents)[indices]
        top_perc = np.array(percs)[indices]
        return top_agents,top_perc
    
    def generate(self):
        #instantate invdivual class
        Individual  = type('Individual', (object,), {})
        self.pop = [Individual() for i in range(self.nPop)]
        
        # best particle   
        self.BestSol = Individual()
        if self.weight == 1:
            self.BestSol.fit = -np.inf
        elif self.weight == -1:
            self.BestSol.fit = np.inf
            
        # initialize an empty newsol
        NewSol = Individual()     
        
        # setting up a variable to store all agents 
        all_agents = []                
        
        # array to save best fits
        self.best_fits = np.zeros(self.MaxIt)            
            
        # initialize initial population
        for i in range(self.nPop):
            self.pop[i].position = np.random.uniform(self.VarMin,self.VarMax,self.VarSize)           
        
            # evaluation
            if (self.x_test is not None) and (self.y_test is not None):
                self.pop[i].fit, self.pop[i].perc = self.fitness.calculate_fitness(self.estimator,
                        self.cv,
                        self.scorer,
                        self.x[:,self.RK(self.pop[i].position)],
                        self.y,
                        x_test=self.x_test[:,self.RK(self.pop[i].position)],
                        y_test=self.y_test)
                
            else:
                self.pop[i].fit, self.pop[i].perc = self.fitness.calculate_fitness(self.estimator,
                        self.cv,
                        self.scorer,
                        self.x[:,self.RK(self.pop[i].position)],
                        self.y)           
        
            #update global best
            if self.weight == 1 and self.pop[i].fit > self.BestSol.fit:
                self.BestSol.position = self.pop[i].position
                self.BestSol.fit = self.pop[i].fit
                self.BestSol.perc = self.pop[i].perc
                
            elif self.weight == -1 and self.pop[i].fit < self.BestSol.fit:
                self.BestSol.position = self.pop[i].position
                self.BestSol.fit = self.pop[i].fit
                self.BestSol.perc = self.pop[i].perc
        
        
        #DE main loop
        for it in range(self.MaxIt):
            for i in range(self.nPop):
                
                ind1 = self.pop[i].position
                
                A = np.random.permutation(self.nPop)
                to_del = np.where(A==i)[0][0]
                A = np.delete(A,to_del)
                
                a = A[0]
                b = A[1]
                c = A[2]
                
                # perform mutation and get a new ind
                ind2 = self.Mutation(a,b,c)
                
                # perform crossover between ind1 and ind2 and get ind3
                ind3 = self.CrossOver(ind1,ind2)
                
                NewSol.position = ind3
                
                # evaluation on ind3
                if (self.x_test is not None) and (self.y_test is not None):
                    NewSol.fit, NewSol.perc = self.fitness.calculate_fitness(self.estimator,
                                                                self.cv,
                                                                self.scorer,
                                                                self.x[:,self.RK(NewSol.position)],
                                                                self.y,
                                                                x_test=self.x_test[:,self.RK(NewSol.position)],
                                                                y_test=self.y_test)   
                    
                else:
                    NewSol.fit, NewSol.perc = self.fitness.calculate_fitness(self.estimator,
                                                                self.cv,
                                                                self.scorer,
                                                                self.x[:,self.RK(NewSol.position)],
                                                                self.y)
                
                # update pop if newsol is better than ith indivdual
                if self.weight == 1 and NewSol.fit > self.pop[i].fit:
                    self.pop[i].position = NewSol.position
                    self.pop[i].fit = NewSol.fit
                    self.pop[i].perc = NewSol.perc
                    
                    #update global best
                    if self.pop[i].fit > self.BestSol.fit:
                        self.BestSol.position = self.pop[i].position
                        self.BestSol.fit = self.pop[i].fit                
                        self.BestSol.perc = self.pop[i].perc                
                    
                elif self.weight == -1 and self.pop[i].fit < self.pop[i].fit:
                    self.pop[i].position = NewSol.position
                    self.pop[i].fit = NewSol.fit
                    self.pop[i].perc = NewSol.perc
                    
                    #update global best
                    if self.pop[i].fit < self.BestSol.fit:
                        self.BestSol.position = self.pop[i].position
                        self.BestSol.fit = self.pop[i].fit
                        self.BestSol.perc = self.pop[i].perc
                
            self.best_fits[it] = self.BestSol.fit 
            
            # updating all agents
            all_agents.extend(self.pop)            
            
            if self.verbose:
                print("Iter: {} Features: {}  Fitness_score: {:.4} ".format(it+1,self.RK(self.BestSol.position),self.BestSol.fit))
        
        # getting top_agents
        tops = self.FeatureRanker(all_agents) 
        
        return np.array([self.RK(self.BestSol.position), self.BestSol.fit]), self.best_fits, tops
                 
    
class DE(BaseFeatureSelector):
    def __init__(self, model, n_f, weight, scoring, n_iter = 1000, n_pop=20, cv=None, 
                 cross_proba = 0.3, low_bound = 0.5, upper_bound = 1, 
                 verbose= True, random_state=None,**kwargs):
        
        """
        DE or Differential Evolution is an global optimization algorithm 
        which was first introduced by Storn and Price in 1990s. 
        DE is a gradient free algorithm and does not require its objective
        to be differentiable. These properties have made DE a great candidate
        for the feature or variable selection problem.
        
        Parameters
        ----------
        model : class
            Instantiated Sklearn regression or classification estimator.
        n_f : int
            Number of features needed to be extracted.
        weight : int
            Maximization or minimization objective, for maximization use +1
            and for mimization use -1.
        scoring : callable
            Callable sklearn score or customized score function that takes in y_pred and y_true                                                
        n_iter : int, optional
            Maximum number of iterations. For more complex 
            problems it is better to set this parameter larger. 
            The default is 1000.
        n_pop : int, optional
            number of population size at each iteration. Typically, this 
            parameter is set to be 10*n_f, but it is dependent on the complexity 
            of the model and it is advised that user tune this parameter based 
            on their problem. The default is 20.
        cv : class, optional
            Instantiated sklearn cross-validation class. The default is None.
        cross_proba : float, optional
            The probability of individuals or agents to mate and generate
            offspring. This value is usually chosen considerably lower than one
            (e.g., 0.3), if convergence is not being achieved one can choose 
            a higher value. The higher the value, the less change of falling
            into local optima but higher compuational time. The lower the
            value the higher chance of falling in local optima but lower
            computational time. The default is 0.3.
        low_bound : float, optional
            The lower bound for choosing beta (F) value in mutation. 
            The default is 0.5.
        upper_bound : float, optional
            The upper bound for choosing beta (F) value in mutation. 
            The default is 1.
        verbose : bool, optional
            Wether to print out progress messages. The default is True.
        random_state : int, optional
            Determines the random state for random number generation, 
            cross-validation, and classifier/regression model setting. 
            The default is None.

        Returns
        -------
        Instantiated Optimziation model class.
        
        References:
            Storn, R. (1996, June). On the usage of differential evolution for 
            function optimization. In Proceedings of North American Fuzzy 
            Information Processing (pp. 519-523). IEEE.
            
        
        """            
        super().__init__(scoring=scoring, n_f = n_f,**kwargs)
        
        # checking the input arguments
        if n_f != int(n_f):
            raise ValueError("n_f should be an integer")
        if n_pop != int(n_pop):
            raise ValueError("n_pop should be an integer")
        if n_iter != int(n_iter):
            raise ValueError("n_iter should be an integer")
        if weight != int(weight):
            raise ValueError("wieght should be a tuple of either -1 or +1")
        try:
            self.fitness = ff
        except:
            raise ImportError("Cannot find/import ff.py defined")
                
        
        self.model = model
        self.n_f = n_f
        self.weight = weight     
        self.scoring = scoring 
        self.n_pop = n_pop
        self.n_iter = n_iter    
        self.cv = cv      
        self.low_bound = low_bound
        self.upper_bound = upper_bound
        self.cross_proba = cross_proba
        
        self.verbose = verbose
        self.random_state = random_state
        
        try:
            self.cv.random_state = self.random_state
            self.cv.shuffle = True
        except:
            pass
        
        try:
            self.model.random_state = self.random_state
        except:
            pass
        

    def fit(self,x,y,decor=None,scale=False, test_size = None, n_repeats = 30,**kwargs):  
        """
        Fit the model to input data X and target values Y, with extra optional 
        arguments for preprocessing data.
        
        Parameters
        ----------
        x : ndarray or sparse matrix of shape (n_samples, n_features)
            Input data.
        y : ndarray of shape (n_samples,)
            Target values.
        decor : tuple or float, optional
            Decorrellation parameter. If a number in range [0,1] is chosen
            then the input features that have a correllation factor above the 
            give value, are removed. If tuple of (string, float [0,1]) is chosen
            then the input data features are ranked based on their feature
            importance method, indicated by the string, then decorrelated and the
            ones with higher importance are retained. The string could be one
            from the following:                
                "variance"
                "f_test"
                "mutual_info"
                "chi2"
                "pearson"                
            The default is None.
        scale : bool, optional
            Wether to scale the input data or not (centering and scaling).
            The default is False.
        test_size : float, optional
            Represents the portion of input data to be included for test split. 
            Needs to be between 0 and 1. The default is None.
        n_repeats : int, optional
            Number of times to permute the feature for the calculation of 
            permutation importance. The default is 30.
            
        Attributes
        ----------
        best_fits: Vector of shape (n_iter,)
            Fitness values of each iteration.        
        best_sol: Vector of shape (n_f,)
            Indices of selected fetures after preprocessing input data.
        best_sol_acc: float
            Optimized fitness value for the identified solution.
        model_best: Class
            Regression/classification estimator given to the optimization 
            algorithm. At this stage model_best = model
        rankings: rankings vectors from the method proposed by Hassanzadeh et al. (2021)            

        Returns
        -------
        Vector of shape (n_f,)
            Indices of selected fetures of the original input data (prior to 
            preprocessing). If no deocr is given, this equals to best_sol.
        Vector of shape (n_iter,)
            Fitness values of each iteration.

        """
        self.load(x,y,decor=decor,scale=scale)        
        if test_size is not None:
            self.train_test_split(test_size, random_state=self.random_state,**kwargs)
                        
        a = time.time()
        if hasattr(self, 'x_train') and hasattr(self, 'x_test') and hasattr(self, 'y_train') and hasattr(self, 'y_test'):
            self.alg_opt = _DifferentialEvolution(self.model,
                                      self.cv,
                                      self.scorer,                                      
                                      self.n_f,
                                      self.fitness,
                                      self.weight,
                                      self.x_train,
                                      self.y_train,                                                                             
                                      x_test = self.x_test,
                                      y_test = self.y_test,
                                      MaxIt = self.n_iter, 
                                      nPop = self.n_pop,
                                      low_bound = self.low_bound, 
                                      upper_bound = self.upper_bound, 
                                      cross_proba = self.cross_proba,
                                      verbose = self.verbose,
                                      random_state = self.random_state)     
            
        else:
            self.alg_opt = _DifferentialEvolution(self.model,
                                      self.cv,
                                      self.scorer,                                      
                                      self.n_f,
                                      self.fitness,
                                      self.weight,
                                      self.x,
                                      self.y,
                                      MaxIt = self.n_iter, 
                                      nPop = self.n_pop,
                                      low_bound = self.low_bound, 
                                      upper_bound = self.upper_bound, 
                                      cross_proba = self.cross_proba,
                                      verbose = self.verbose,
                                      random_state = self.random_state)     


        best_sol, self.best_fits, tops = self.alg_opt.generate()
        self.best_sol = best_sol[0]
        self.best_sol_acc = best_sol[1]       
        self.best_features = self.best_sol
        # setting the model_best = model since one may not want to run tune_param
        self.model_best = clone(self.model)
        
        #setting top rankings 
        self.rankings = self._feature_ranking(tops, self.x_cols, n_repeats=n_repeats, random_state= self.random_state)
        
        if self.verbose:
            print("Optimization completed in {} seconds".format(time.time() - a))
            print("Best individual is {}, {:.4}".format(np.array(self.x_cols[self.best_sol]), self.best_sol_acc))
        
        return self.x_cols[self.best_sol], self.best_fits  
    
    @property
    def _name_(self):
        return "DE"
                
                
                
                
                
                
                
                
                
                
        


from setuptools import find_packages, setup
import os
base_dir = os.path.dirname(__file__)

with open(os.path.join(base_dir, "README.md")) as f:
    long_description = f.read()

    
setup(
	name= "jostar",
	version="0.0.3",
	description="Feature Selection Module for Data Sciences in Python",
	long_description=long_description,
	long_description_content_type = "text/markdown",
	author="Amirhossein Hassanzadeh",
	url = "",
	author_email = "ah7557@rit.edu" ,
	packages=find_packages(),
	keywords={"optimization", "feature-selection","genetic algorithm", "particle swarm", "ant colony", "metaheuristics", "differential evolution", "sequential search", "multi-objective"},
	classifiers=['Programming Language :: Python :: 3.6',
				"License :: OSI Approved :: MIT License",
				"Operating System :: OS Independent"],
	install_requires=["sklearn","matplotlib","pathos","scipy","seaborn"],
	extras_require={"dev":["pytest>=3.6",],}
	)
	

